# -*- coding: utf-8 -*-

"""Tests for :mod:`~utilipy.utils.inspect`."""


__all__ = [
    "test_parameters",
    "test__is_empty",
    "test__is_void",
    "test__is_placehold",
    "test__is_placeholder",
    "test_getfullerargspec",
    "test_get_annotations_from_signature",
    "test_get_defaults_from_signature",
    "test_get_kwdefaults_from_signature",
    "test_get_kwonlydefaults_from_signature",
    "test_get_kinds_from_signature",
    "test_modify_parameter",
    "test_replace_with_parameter",
    "test_insert_parameter",
    "test_prepend_parameter",
    "test_append_parameter",
    "test_drop_parameter",
    # test FullerSignature
    "test_FullerSignature_creation",
    "test_FullerSignature_bind",
    "test_FullerSignature_signature",
    "test_FullerSignature_annotations",
    "test_FullerSignature_defaults",
    "test_FullerSignature_kwdefaults",
    "test_FullerSignature_kinds",
    "test_FullerSignature_names",
    "test_FullerSignature_index_positional",
    "test_FullerSignature_index_positional_only",
    "test_FullerSignature_defaulted",
    "test_FullerSignature_index_var_positional",
    "test_FullerSignature_index_keyword_only",
    "test_FullerSignature_index_end_keyword_only",
    "test_FullerSignature_index_var_keyword",
    "test_FullerSignature_copy",
    "test_FullerSignature_modify_parameter",
    "test_FullerSignature_replace_with_parameter",
    "test_FullerSignature_insert_parameter",
    "test_FullerSignature_prepend_parameter",
    "test_FullerSignature_append_parameter",
    "test_FullerSignature_drop_parameter",
    "test_FullerSignature_add_var_positional_parameter",
    "test_FullerSignature_add_var_keyword_parameter",
    # "test_FullerSignature_hidden_methods",
    # fuller_signature
    "test_fuller_signature",
    "test_signature_from_method",
    "test_fuller_signature_from_method",
]


##############################################################################
# IMPORTS

# BUILT-IN
import inspect as nspct
from collections import OrderedDict
from types import MappingProxyType

# THIRD PARTY
import pytest

# PROJECT-SPECIFIC
from utilipy.utils import inspect

##############################################################################
# PARAMETERS


class NS:
    """Name Space. Testing without messing up pytest."""

    @staticmethod
    def test_func(
        x: int, y, a: int = 1, b=2, *args: str, j: str = "a", k="b", **kw: dict
    ) -> bool:
        """Function with all possible arguments."""
        x, y, a, b, args, j, k, kw  # pragma: no cover
        return True

    # /def

    @staticmethod
    def appendable_func(x, y):
        """Fumction with only positional-or-keyword arguments."""

    # /def

    @staticmethod
    def no_arg_func():
        """Function with no arguments."""

    # /def

    class ExceptEquality(object):
        """Throw an Exception on any equality test."""

        def __eq__(self, other):
            """No Equality."""
            raise Exception(f"forbidden equality tests with {other}")

        # /def

    # /class

    def method(
        self,
        x: int,
        y,
        a: int = 1,
        b=2,
        *args: str,
        j: str = "a",
        k="b",
        **kw: dict,
    ):
        pass

    # /def

    @classmethod
    def classmethod(
        cls,
        x: int,
        y,
        a: int = 1,
        b=2,
        *args: str,
        j: str = "a",
        k="b",
        **kw: dict,
    ):
        pass

    # /def


# /class

NS.NEQ = NS.ExceptEquality()
NS.normsig = nspct.Signature.from_callable(NS.test_func)
NS.fullsig = inspect.FullerSignature.from_callable(NS.test_func)
NS.noargsig = inspect.FullerSignature.from_callable(NS.no_arg_func)
NS.appsig = inspect.FullerSignature.from_callable(NS.appendable_func)


##############################################################################
# Tests
##############################################################################


def test_parameters():
    """Test array_like.

    TODO

    """
    # types
    assert inspect.POSITIONAL_ONLY == nspct.Parameter.POSITIONAL_ONLY
    assert (
        inspect.POSITIONAL_OR_KEYWORD == nspct.Parameter.POSITIONAL_OR_KEYWORD
    )
    assert inspect.VAR_POSITIONAL == nspct.Parameter.VAR_POSITIONAL
    assert inspect.KEYWORD_ONLY == nspct.Parameter.KEYWORD_ONLY
    assert inspect.VAR_KEYWORD == nspct.Parameter.VAR_KEYWORD

    # placeholders
    assert inspect._empty == nspct.Parameter.empty
    assert inspect._void == nspct._void

    assert type(inspect._placehold()) == inspect._placehold

    assert issubclass(inspect.FullerArgSpec, tuple)  # TODO full type check
    FAS = inspect.FullerArgSpec(1, 2, 3, 4, 5, 6, 7, 8, 9)
    assert all(
        (
            FAS.args == 1,
            FAS.defaultargs == 2,
            FAS.argdefaults == 3,
            FAS.varargs == 4,
            FAS.kwonlyargs == 5,
            FAS.kwonlydefaults == 6,
            FAS.varkw == 7,
            FAS.annotations == 8,
            FAS.docstring == 9,
        )
    )


# /def


##########################################################################
# safe placeholder comparison


def test__is_empty():
    """Test _is_empty."""
    # True
    assert inspect._is_empty(inspect._empty)  # my inspect module
    assert inspect._is_empty(nspct.Parameter.empty)  # built-in

    # False
    assert not inspect._is_empty(1)

    # Exception
    assert not inspect._is_empty(NS.NEQ)


# /def


def test__is_void():
    """Test _is_void."""

    # True
    assert inspect._is_void(inspect._void)  # my inspect module
    assert inspect._is_void(nspct._void)  # built-in

    # False
    assert not inspect._is_void(1)

    # Exception
    assert not inspect._is_void(NS.NEQ)


# /def


def test__is_placehold():
    """Test _is_placehold."""

    # True
    assert inspect._is_placehold(inspect._placehold)  # my inspect module

    # False
    assert not inspect._is_placehold(1)

    # Exception
    assert not inspect._is_placehold(NS.NEQ)


# /def


def test__is_placeholder():
    """Test _is_placeholder."""
    # True
    assert all(
        (
            inspect._is_placeholder(inspect._empty),
            inspect._is_placeholder(inspect._void),
            inspect._is_placeholder(inspect._placehold),
        )
    )

    # False
    assert not inspect._is_placeholder(1)

    # Exception
    assert not inspect._is_placeholder(NS.NEQ)


# /def


###########################################################################
# getfullerargspec


def test_getfullerargspec():
    """Test :func:`~utilipy.utils.inspect.getfullerargspec`"""
    fullargspec = nspct.getfullargspec(NS.test_func)
    fullerargspec = inspect.getfullerargspec(NS.test_func)

    # arguments
    assert fullerargspec.args == ["x", "y"] == fullargspec.args[:-2]

    # default arguments
    assert fullerargspec.defaultargs == ["a", "b"] == fullargspec.args[-2:]
    assert fullerargspec.argdefaults == {"a": 1, "b": 2}
    assert tuple(fullerargspec.argdefaults.values()) == fullargspec.defaults

    # varargs
    assert fullerargspec.varargs == fullargspec.varargs

    # kwonlyargs
    assert fullerargspec.kwonlyargs == fullargspec.kwonlyargs
    assert fullerargspec.kwonlydefaults == fullargspec.kwonlydefaults

    # varkw
    assert fullerargspec.varkw == fullargspec.varkw

    # annotations
    assert fullerargspec.annotations == fullargspec.annotations

    assert fullerargspec.docstring == NS.test_func.__doc__

    return


# /def


###########################################################################
# Signature / ArgSpec Interface


def test_get_annotations_from_signature():
    """Test get_annotations_from_signature."""
    signature = inspect.signature(NS.test_func)
    annotations = inspect.get_annotations_from_signature(signature)

    assert annotations == {
        "x": int,
        "a": int,
        "args": str,
        "j": str,
        "kw": dict,
        "return": bool,
    }

    return


# /def


def test_get_defaults_from_signature():
    """Test get_annotations_from_signature."""
    signature = inspect.signature(NS.test_func)
    defaults = inspect.get_defaults_from_signature(signature)

    assert defaults == (1, 2)

    return


# /def


def test_get_kwdefaults_from_signature():
    """Test get_annotations_from_signature."""
    signature = inspect.signature(NS.test_func)
    kwdefaults = inspect.get_kwdefaults_from_signature(signature)

    assert kwdefaults == {"j": "a", "k": "b"}

    return


# /def


def test_get_kwonlydefaults_from_signature():
    """Test get_annotations_from_signature."""
    signature = inspect.signature(NS.test_func)
    kwdefaults = inspect.get_kwonlydefaults_from_signature(signature)

    assert kwdefaults == {"j": "a", "k": "b"}

    return


# /def


def test_get_kinds_from_signature():
    """Test get_annotations_from_signature."""
    signature = inspect.signature(NS.test_func)
    kinds = inspect.get_kinds_from_signature(signature)

    assert kinds == (
        inspect.POSITIONAL_OR_KEYWORD,
        inspect.POSITIONAL_OR_KEYWORD,
        inspect.POSITIONAL_OR_KEYWORD,
        inspect.POSITIONAL_OR_KEYWORD,
        inspect.VAR_POSITIONAL,
        inspect.KEYWORD_ONLY,
        inspect.KEYWORD_ONLY,
        inspect.VAR_KEYWORD,
    )


###########################################################################
# Signature Methods


def test_modify_parameter():
    """Test modify_parameter."""
    signature = inspect.signature(NS.test_func)

    # ----------------------------------------------------
    # doing piecemeal for full code coverage
    # and so can cycle through index / versus name as `param` arg
    sig = inspect.modify_parameter(signature, "x", name="xx")
    sig = inspect.modify_parameter(sig, "xx", kind=inspect.POSITIONAL_ONLY)
    sig = inspect.modify_parameter(sig, "y", default="Y")
    sig = inspect.modify_parameter(sig, 1, annotation="YY")

    parameters = list(sig.parameters.values())

    # changed
    assert parameters[0].name == "xx"
    assert parameters[0].kind == inspect.POSITIONAL_ONLY
    assert parameters[1].default == "Y"
    assert parameters[1].annotation == "YY"

    # not changed
    assert parameters[0].annotation == int
    assert parameters[1].name == "y"

    return


# /def


def test_replace_with_parameter():
    """Test replace_with_parameter."""
    signature = inspect.signature(NS.test_func)

    sig = inspect.modify_parameter(signature, 0, name="xx")
    sig = inspect.modify_parameter(sig, 1, name="yy")
    new_param0 = list(sig.parameters.values())[0]
    new_param1 = list(sig.parameters.values())[1]

    signature = inspect.replace_with_parameter(signature, "x", new_param0)
    signature = inspect.replace_with_parameter(signature, 1, new_param1)

    parameters = list(sig.parameters.values())

    assert parameters[0].name == "xx"
    assert parameters[1].name == "yy"

    return


# /def


def test_insert_parameter():
    """Test insert_parameter."""
    signature = inspect.signature(NS.test_func)

    sig = inspect.modify_parameter(signature, 0, name="xx")
    new_param = list(sig.parameters.values())[0]

    signature = inspect.insert_parameter(signature, 1, new_param)

    assert list(signature.parameters.values())[1].name == "xx"

    return


# /def


def test_prepend_parameter():
    """Test prepend_parameter."""
    signature = inspect.signature(NS.test_func)

    sig = inspect.modify_parameter(signature, 0, name="xx")
    new_param = list(sig.parameters.values())[0]

    signature = inspect.prepend_parameter(signature, new_param)

    assert list(signature.parameters.values())[0].name == "xx"

    return


# /def


def test_append_parameter():
    """Test append_parameter."""
    signature = inspect.signature(NS.appendable_func)

    sig = inspect.modify_parameter(signature, 0, name="xx")
    new_param = list(sig.parameters.values())[0]

    signature = inspect.append_parameter(signature, new_param)

    assert list(signature.parameters.values())[2].name == "xx"

    return


# /def


def test_drop_parameter():
    """Test drop_parameter."""
    signature = inspect.signature(NS.test_func)

    # don't drop anything
    newsignature = inspect.drop_parameter(signature, None)
    assert newsignature == signature

    # drop x, by name
    signature = inspect.drop_parameter(signature, "x")
    assert list(signature.parameters.values())[0].name == "y"

    # drop y, by index
    signature = inspect.drop_parameter(signature, 0)
    assert list(signature.parameters.values())[0].name == "a"

    # drop a, by parameter
    signature = inspect.drop_parameter(
        signature, list(signature.parameters.values())[0]
    )
    assert list(signature.parameters.values())[0].name == "b"

    # exception
    with pytest.raises(TypeError):  # can't drop a set
        inspect.drop_parameter(signature, set())

    return


# /def


###########################################################################
# Signature


def test_FullerSignature_creation():
    """Test `~utilipy.utils.inspect.FullerSignature` class creation."""
    # ------------------
    # __init__

    blanksig = inspect.FullerSignature()

    assert blanksig.parameters == MappingProxyType(OrderedDict())
    assert blanksig.return_annotation == inspect._empty
    assert blanksig.obj is None

    # manual construction
    parameters = {
        "x": inspect.Parameter("x", inspect.POSITIONAL_OR_KEYWORD, default=0),
        "y": inspect.Parameter("y", inspect.POSITIONAL_OR_KEYWORD, default=1),
    }

    madesig = inspect.FullerSignature(
        parameters=parameters.values(), return_annotation="return_"
    )

    assert madesig.parameters == parameters
    assert madesig.return_annotation == "return_"
    assert madesig.obj is None

    # ------------------
    # from_callable

    normsig = nspct.Signature.from_callable(NS.test_func)
    fromcallsig = inspect.FullerSignature.from_callable(NS.test_func)

    assert fromcallsig.parameters == normsig.parameters
    assert fromcallsig.return_annotation == normsig.return_annotation
    assert fromcallsig.obj == NS.test_func

    # ------------------
    # from_signature

    fromsigsig = inspect.FullerSignature.from_signature(
        fromcallsig, obj=NS.test_func
    )

    assert fromsigsig.parameters == fromcallsig.parameters
    assert fromsigsig.return_annotation == fromcallsig.return_annotation
    assert fromsigsig.obj == NS.test_func


# /def


def test_FullerSignature_bind():
    """Test `~utilipy.utils.inspect.FullerSignature` argument binding.

    The ``bind`` and ``bind_partial`` methods are tested by base python.
    Here we test the ``bind_with_defaults`` and ``bind_partial_with_defaults``

    """
    # ------------------
    # bind_with_defaults

    sig = NS.fullsig

    ba = sig.bind_with_defaults(-1, -2, 1.1, 2.2, 4, 5, l="cc")

    assert ba.args == (-1, -2, 1.1, 2.2, 4, 5)
    assert ba.kwargs == dict(j="a", k="b", l="cc")  # defaults in here

    # ------------------
    # bind_partial_with_defaults

    ba = sig.bind_partial_with_defaults(-1, l="cc")

    assert ba.args == (-1,)  # no y
    assert ba.arguments["a"] == 1
    assert ba.arguments["b"] == 2
    assert ba.kwargs == dict(
        a=1, b=2, args=(), j="a", k="b", l="cc"
    )  # defaults in here


# /def


def test_FullerSignature_signature():
    """Test `~utilipy.utils.inspect.FullerSignature` signature."""
    normsig = NS.normsig
    fullsig = NS.fullsig

    assert fullsig.signature == fullsig.__signature__ == normsig


# /def


def test_FullerSignature_annotations():
    normsig = NS.normsig
    fullsig = NS.fullsig

    assert (
        fullsig.annotations
        == fullsig.__annotations__
        == inspect.get_annotations_from_signature(normsig)
    )


# /def


def test_FullerSignature_defaults():
    normsig = NS.normsig
    fullsig = NS.fullsig

    assert (
        fullsig.defaults
        == fullsig.__defaults__
        == inspect.get_defaults_from_signature(normsig)
    )


# /def


def test_FullerSignature_kwdefaults():
    normsig = NS.normsig
    fullsig = NS.fullsig

    assert (
        fullsig.kwdefaults
        == fullsig.__kwdefaults__
        == fullsig.kwonlydefaults
        == inspect.get_kwdefaults_from_signature(normsig)
    )


# /def


def test_FullerSignature_kinds():
    normsig = NS.normsig
    fullsig = NS.fullsig

    assert fullsig.kinds == inspect.get_kinds_from_signature(normsig)


# /def


def test_FullerSignature_names():

    assert NS.fullsig.names == ("x", "y", "a", "b", "args", "j", "k", "kw")


# /def


def test_FullerSignature_index_positional():
    noargsig = NS.noargsig
    fullsig = NS.fullsig

    assert fullsig.index_positional == (0, 1, 2, 3)
    assert noargsig.index_positional is False


# /def


def test_FullerSignature_index_positional_only():
    fullsig = NS.fullsig

    assert fullsig.index_positional_only is False
    # need to make such a signature
    sig = fullsig.modify_parameter("x", kind=inspect.POSITIONAL_ONLY)
    assert sig.index_positional_only == (0,)


# /def


def test_FullerSignature_defaulted():
    fullsig = NS.fullsig
    noargsig = NS.noargsig

    assert fullsig.index_positional_defaulted == (0, 1, 2, 3)  # FIXME
    assert noargsig.index_positional_defaulted is False


# /def


def test_FullerSignature_index_var_positional():
    fullsig = NS.fullsig
    noargsig = NS.noargsig

    assert fullsig.index_var_positional == 4
    assert noargsig.index_var_positional is False


# /def


def test_FullerSignature_index_keyword_only():
    fullsig = NS.fullsig
    noargsig = NS.noargsig

    assert fullsig.index_keyword_only == (5, 6)
    assert noargsig.index_keyword_only is False


# /def


def test_FullerSignature_index_end_keyword_only():
    fullsig = NS.fullsig
    noargsig = NS.noargsig

    assert fullsig.index_end_keyword_only == (6 + 1)
    assert noargsig.index_end_keyword_only == (0 + 1)


# /def


def test_FullerSignature_index_var_keyword():
    fullsig = NS.fullsig
    noargsig = NS.noargsig

    assert fullsig.index_var_keyword == 7
    assert noargsig.index_var_keyword is False


# /def


def test_FullerSignature_copy():
    fullsig = NS.fullsig
    noargsig = NS.noargsig

    assert fullsig == fullsig.copy()
    assert noargsig == noargsig.copy()


# /def


def test_FullerSignature_modify_parameter():
    fullsig = NS.fullsig

    # doing piecemeal for full code coverage
    # and so can cycle through index / versus name as `param` arg
    sig = fullsig.modify_parameter("x", name="xx")
    sig = sig.modify_parameter("xx", kind=inspect.POSITIONAL_ONLY)
    sig = sig.modify_parameter("y", default="Y")
    sig = sig.modify_parameter(1, annotation="YY")

    assert isinstance(sig, inspect.FullerSignature)

    parameters = list(sig.parameters.values())

    # changed
    assert parameters[0].name == "xx"
    assert parameters[0].kind == inspect.POSITIONAL_ONLY
    assert parameters[1].default == "Y"
    assert parameters[1].annotation == "YY"

    # not changed
    assert parameters[0].annotation == int
    assert parameters[1].name == "y"


# /def


def test_FullerSignature_replace_with_parameter():
    fullsig = NS.fullsig

    sig = fullsig.modify_parameter(0, name="xx")
    sig = sig.modify_parameter(1, name="yy")
    new_param0 = list(sig.parameters.values())[0]
    new_param1 = list(sig.parameters.values())[1]

    sig = fullsig.replace_with_parameter("x", new_param0)
    sig = sig.replace_with_parameter(1, new_param1)

    assert isinstance(sig, inspect.FullerSignature)

    parameters = list(sig.parameters.values())

    assert parameters[0].name == "xx"
    assert parameters[1].name == "yy"


# /def


def test_FullerSignature_insert_parameter():
    fullsig = NS.fullsig

    sig = fullsig.modify_parameter(0, name="xx")
    new_param = list(sig.parameters.values())[0]

    sig = fullsig.insert_parameter(1, new_param)

    assert isinstance(sig, inspect.FullerSignature)

    assert list(sig.parameters.values())[1].name == "xx"


# /def


def test_FullerSignature_prepend_parameter():
    fullsig = NS.fullsig

    sig = fullsig.modify_parameter(0, name="xx")
    new_param = list(sig.parameters.values())[0]

    sig = fullsig.prepend_parameter(new_param)

    assert isinstance(sig, inspect.FullerSignature)

    assert list(sig.parameters.values())[0].name == "xx"


# /def


def test_FullerSignature_append_parameter():
    appsig = NS.appsig

    sig = appsig.modify_parameter(0, name="xx")
    new_param = list(sig.parameters.values())[0]

    sig = appsig.append_parameter(new_param)

    assert isinstance(sig, inspect.FullerSignature)

    assert list(sig.parameters.values())[2].name == "xx"


# /def


def test_FullerSignature_drop_parameter():
    fullsig = NS.fullsig

    # drop x, by name
    sig = fullsig.drop_parameter("x")
    assert list(sig.parameters.values())[0].name == "y"

    # drop y, by index
    sig = sig.drop_parameter(0)
    assert list(sig.parameters.values())[0].name == "a"

    # drop a, by parameter
    sig = sig.drop_parameter(list(sig.parameters.values())[0])
    assert list(sig.parameters.values())[0].name == "b"


# /def


def test_FullerSignature_add_var_positional_parameter():
    fullsig = NS.fullsig
    noargsig = NS.noargsig

    sig1 = fullsig.add_var_positional_parameter()
    sig2 = fullsig.add_var_positional_parameter(name="nargs")  # not changed
    sig3 = noargsig.add_var_positional_parameter(name="nargs")

    assert list(sig1.parameters.values())[4].name == "args"
    assert list(sig2.parameters.values())[4].name == "args"
    assert list(sig3.parameters.values())[0].name == "nargs"

    # if index in self.index_positional_defaulted
    dropsig = fullsig.drop_parameter("args")
    sig4 = dropsig.add_var_positional_parameter(name="args", index=3)
    assert list(sig4.parameters.values())[3].name == "args"

    # index < self.index_positional_defaulted[0]
    sig5 = dropsig.add_var_positional_parameter(name="args", index=2)
    assert list(sig5.parameters.values())[2].name == "args"

    # adding what already exists
    sig6 = dropsig.add_var_positional_parameter()
    assert list(sig6.parameters.values())[4].name == "args"


# /def


def test_FullerSignature_add_var_keyword_parameter():
    fullsig = NS.fullsig
    noargsig = NS.noargsig

    sig1 = fullsig.add_var_keyword_parameter()
    sig2 = fullsig.add_var_keyword_parameter(name="nargs")  # not changed
    sig3 = noargsig.add_var_keyword_parameter(name="kw")

    assert list(sig1.parameters.values())[-1].name == "kw"
    assert list(sig2.parameters.values())[-1].name == "kw"
    assert list(sig3.parameters.values())[-1].name == "kw"

    return


# /def


# def test_FullerSignature_hidden_methods():
#     fullsig = NS.fullsig

#     sig = fullsig._default_pos_to_kwonly_from(index=3)

#     assert False, list(sig.parameters.values())[2].kind


# # /def

# ------------------------------------------------------------------------


def test_fuller_signature():
    """Test `~utilipy.utils.inspect.fuller_signature`."""
    test_FullerSignature_creation()


# /def


def test_signature_from_method():
    """Test `~utilipy.utils.inspect.signature_from_method`."""
    # instance method
    sig = inspect.signature_from_method(NS.method)

    assert tuple(sig.parameters.values()) == tuple(
        NS.fullsig.signature.parameters.values()
    )

    # class method

    sig = inspect.signature_from_method(NS.classmethod)

    assert tuple(sig.parameters.values()) == tuple(
        NS.fullsig.signature.parameters.values()
    )


# /def


def test_fuller_signature_from_method():
    """Test `~utilipy.utils.inspect.fuller_signature_from_method`."""
    # instance method
    sig = inspect.fuller_signature_from_method(NS.method)

    assert tuple(sig.parameters.values()) == tuple(
        NS.fullsig.parameters.values()
    )

    # class method

    sig = inspect.fuller_signature_from_method(NS.classmethod)

    assert tuple(sig.parameters.values()) == tuple(
        NS.fullsig.parameters.values()
    )


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : astronat imports
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

"""Base set of imports for :mod:`~astronat`.

Only imports if astronat is installed

Routine Listings
----------------
astronat : imports

    - constants, dynamics, units, utils
    - reload_config -> astronat_reload_config
    - online_help -> astronat_online_help
    - help -> astronat_help

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "astronat_imports_help",
    # modules
    "astronat",
    "constants",
    "dynamics",
    "units",
    "utils",
    # functions
    "astronat_reload_config",
    "astronat_online_help",
    "astronat_help",
    "TableList",
    "QTableList",
    "TablesList",
]


##############################################################################
# HELPER FUNCTIONS

# PROJECT-SPECIFIC
from utilipy.imports import conf
from utilipy.utils import make_help_function

##############################################################################
# IMPORTS

try:

    import astronat

except ImportError:

    import warnings

    warnings.warn("Cannot import astronat")

else:

    # modules
    from astronat import constants, dynamics, units, utils

    # functions
    from astronat import (
        reload_config as astronat_reload_config,
        online_help as astronat_online_help,
        help as astronat_help,
    )

    from astronat.utils.table import TableList, QTableList, TablesList


# /if


##############################################################################
# Printing Information

astronat_imports_help = make_help_function(
    "astronat", __doc__, look_for="Routine Listings"
)


if conf.verbose_imports:
    astronat_imports_help()


##############################################################################
# END



# -*- coding: utf-8 -*-
# see LICENSE.rst

"""The `utilipy` Package on import.

Welcome to utilipy, a collection of useful python codes. This is a centralized
repository for non project-specific code. There are modules for making
advanced decorators, interfacing with IPython environments, data utilities,
making fitting libraries inter-operable, and much more.


References
----------
Zenodo [1]_, Astropy [2]_

.. [1] nstarman. (2020, March 23). nstarman/utilipy: astropy_template (Version
    astropy_template). Zenodo. http://doi.org/10.5281/zenodo.3724822
.. [2] Astropy Collaboration et al., 2018, AJ, 156, 123.


Examples
--------

To do a specific search of `utilipy`'s docs, use the ``online_help`` function.
If you don't have a specific query, that's fine too,
`utilipy` will open the general search page.

As an example, here we query RTD for the documentation on `LogFile`.

    >>> import utilipy
    >>> utilipy.online_help(query="LogFile") # doctest: +SKIP

The same can be accomplished with the general `help` function.

    >>> import utilipy
    >>> utilipy.help(query="LogFile", online=True) # doctest: +SKIP

"""

__author__ = "Nathaniel Starkman"

__all__ = [
    # modules
    "data_utils",
    "decorators",
    "extern",
    "imports",
    "math",
    "plot",
    "scripts",
    "utils",
    # functions
    "data_graph",
    "LogFile",
    "ObjDict",
    "wraps",
    "lookup",
    "online_help",
    "reload_config",
]

__all_top_imports__ = (  # TODO deprecate
    "data_utils",
    "decorators",
    "extern",
    "imports",
    "math",
    "plot",
    "scripts",
    "utils",
)


##############################################################################
# IMPORTS

# -----------------------------------------------
# keep this content at the top. (sets the __version__)
from ._astropy_init import *  # noqa: F401, F403  # isort:skip
from ._astropy_init import __version__  # noqa  # isort:skip

# -----------------------------------------------

# BUILT-IN
import typing as T

# THIRD PARTY
import astropy.config as config
from astropy.utils.misc import find_api_page

# PROJECT-SPECIFIC
from . import math  # noqa
from . import data_utils, decorators, extern, imports, plot, scripts, utils
from .data_utils import data_graph
from .utils.collections import ObjDict
from .utils.functools import wraps
from .utils.logging import LogFile

#############################################################################
# CONFIG FUNCTIONS


def reload_config():
    """Reload utilipy configuration.

    See Also
    --------
    :mod:`~astropy.config`

    """
    config.reload_config("utilipy")


# /def


#############################################################################
# HELP FUNCTIONS


def online_help(query: T.Union[None, str, T.Any] = None, version=__version__):
    """Search the online documentation for the given query.

    Opens the results in the default web browser.
    Requires an active internet connection.

    Parameters
    ----------
    query : str, object, optional
        The search query for `RTD <https://utislipy.readthedocs.io>`_.
        None (default) or "" is an empty search.
        If an object, uses :func:`~astropy.utils.misc.find_api_page`
        to find the correct API page.

    Examples
    --------
    To do a specific search of `utilipy`'s docs, use the ``online_help``
    function. If you don't have a specific query, that's fine too,
    `utilipy` will open the general search page.

    As an example, here we query RTD for the documentation on `LogFile`.

        >>> import utilipy
        >>> utilipy.online_help(query="LogFile") # doctest: +SKIP

    The same can be accomplished with the general `help` function.

        >>> import utilipy
        >>> utilipy.help(query="LogFile", online=True) # doctest: +SKIP

    See Also
    --------
    :func:`~utilipy.help`

    """
    # first get version to search
    if "dev" in version:
        version = "latest"
    else:
        version = version if version.startswith("v") else "v" + version

    # look up objects
    if not isinstance(query, str) and query is not None:
        if version == "latest":
            version = None
        return find_api_page(query, version=version, openinbrowser=True)

    # else:

    from urllib.parse import urlencode
    import webbrowser

    # process the query
    if query is None:  # empty query, empty search
        query = ""
    else:  # encode the query
        query: str = urlencode({"q": query})

    url = f"https://utilipy.readthedocs.io/en/{version}/search.html?{query}"

    webbrowser.open(url)


# /def


# -------------------------------------------------------------------


@decorators.code_dev.indev
def lookup(query: T.Optional[str] = None, online: bool = False):
    """*Utilipy* help function. Online search or offline overview.

    Parameters
    ----------
    query : str, optional
        The search query.
    online : bool, optional
        Whether to open the online help or just print some help
        documentation (default)

    Examples
    --------
    A search of `utilipy`'s online docs.

        >>> import utilipy
        >>> utilipy.help(query="LogFile", online=True) # doctest: +SKIP

    See Also
    --------
    :func:`~utilipy.online_help`

    """
    if online:
        return online_help(query=query, version="latest")
    # else:

    print("This function is a work in progress")

    print("".join(["-"] * 79))
    ipython.help()

    return


# /def


#############################################################################
# END



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Astropy documentation build configuration file.

This file is execfile()d with the current directory set to its containing
dir.

Note that not all possible configuration values are present in this file.

All configuration values have a default. Some values are defined in
the global Astropy configuration which is loaded here before anything else.
See astropy.sphinx.conf for which values are set there.

If extensions (or modules to document with autodoc) are in another
directory, add these directories to sys.path here. If the directory is
relative to the documentation root, use os.path.abspath to make it absolute,
like shown here. sys.path.insert(0, os.path.abspath('..')) IMPORTANT: the
above commented section was generated by sphinx-quickstart, but is *NOT*
appropriate for astropy or Astropy affiliated packages. It is left commented
out with this explanation to make it clear why this should not be done. If
the sys.path entry above is added, when the astropy.sphinx.conf import
occurs, it will import the *source* version of astropy instead of the
version installed (if invoked as "make html" or directly with sphinx), or
the version in the build directory (if "python setup.py build_sphinx" is
used). Thus, any C-extensions that are needed to build the documentation
will *not* be accessible, and the documentation will not build correctly.

"""

# BUILT-IN
import datetime
import os
import sys
from configparser import ConfigParser
from importlib import import_module

try:
    from sphinx_astropy.conf.v1 import *  # noqa
except ImportError:
    print("ERROR: the documentation requires the sphinx-astropy package.")
    sys.exit(1)


conf = ConfigParser()

conf.read([os.path.join(os.path.dirname(__file__), "..", "setup.cfg")])
setup_cfg = dict(conf.items("metadata"))

# -- General configuration ----------------------------------------------------

# By default, highlight as Python 3.
highlight_language = "python3"

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.2'

# To perform a Sphinx version check that needs to be more specific than
# major.minor, call `check_sphinx_version("x.y.z")` here.
# check_sphinx_version("1.2.1")

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns.append("_templates")

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog += """
"""

# modindex formatting
modindex_common_prefix = ["utilipy."]

# extensions
extensions += [
    "sphinx.ext.todo",
    # "easydev.copybutton",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinx_automodapi.smart_resolver",
    "sphinxcontrib.bibtex",
]

# Show / hide TODO blocks
todo_include_todos = True

# # Copy Buttton
# jscopybutton_path = easydev.copybutton.get_copybutton_path()
# if os.path.isdir('_static') is False:
#     os.mkdir('_static')
# shutil.copy(jscopybutton_path, '_static')
# html_static_path = ['_static']


# -- Project information ------------------------------------------------------

# This does not *have* to match the package name, but typically does
project = setup_cfg["name"]
author = setup_cfg["author"]
copyright = "{0}, {1}".format(
    datetime.datetime.now().year, setup_cfg["author"]
)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

import_module(setup_cfg["name"])
package = sys.modules[setup_cfg["name"]]

# The short X.Y version.
version = package.__version__.split("-", 1)[0]
# The full version, including alpha/beta/rc tags.
release = package.__version__


# -- Options for HTML output --------------------------------------------------

# A NOTE ON HTML THEMES
# The global astropy configuration uses a custom theme, 'bootstrap-astropy',
# which is installed along with astropy. A different theme can be used or
# the options for this theme can be modified by overriding some of the
# variables set in the global configuration. The variables set in the
# global configuration are listed below, commented out.


# Add any paths that contain custom themes here, relative to this directory.
# To use a different custom theme, add the directory containing the theme.
# html_theme_path = []

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes. To override the custom theme, set this to the
# name of a builtin theme or the name of a custom theme in html_theme_path.
# html_theme = None


html_theme_options = {
    "logotext1": "utilipy",  # white,  semi-bold
    "logotext2": "",  # orange, light
    "logotext3": ":docs",  # white,  light
}


# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = ''

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = ''

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "{0} v{1}".format(project, release)

# Output file base name for HTML help builder.
htmlhelp_basename = project + "doc"


# -- Options for LaTeX output -------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("index", project + ".tex", project + u" Documentation", author, "manual")
]


# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", project.lower(), project + u" Documentation", [author], 1)
]


# -- Options for the edit_on_github extension ---------------------------------

if setup_cfg.get("edit_on_github").lower() == "true":

    extensions += ["sphinx_astropy.ext.edit_on_github"]

    edit_on_github_project = setup_cfg["github_project"]
    edit_on_github_branch = "master"

    edit_on_github_source_root = ""
    edit_on_github_doc_root = "docs"

# -- Resolving issue number to links in changelog -----------------------------
github_issues_url = "https://github.com/{0}/issues/".format(
    setup_cfg["github_project"]
)

# -- Turn on nitpicky mode for sphinx (to warn about references not found) ----
#
# nitpicky = True
# nitpick_ignore = []
#
# Some warnings are impossible to suppress, and you can list specific references
# that should be ignored in a nitpick-exceptions file which should be inside
# the docs/ directory. The format of the file should be:
#
# <type> <class>
#
# for example:
#
# py:class astropy.io.votable.tree.Element
# py:class astropy.io.votable.tree.SimpleElement
# py:class astropy.io.votable.tree.SimpleElementWithContent
#
# Uncomment the following lines to enable the exceptions:
#
# for line in open('nitpick-exceptions'):
#     if line.strip() == "" or line.startswith("#"):
#         continue
#     dtype, target = line.split(None, 1)
#     target = target.strip()
#     nitpick_ignore.append((dtype, six.u(target)))



# -*- coding: utf-8 -*-


"""test functions for base imports."""

__all__ = [
    "test_import_base",
    "test_import_base_has_ipython",
]


##############################################################################


def test_import_base():
    """Test base imports when not in ipython environment."""
    from utilipy.imports import base_imports as imports

    for obj in (
        "os",
        "sys",
        "time",
        "warnings",
        "np",
        "scipy",
        "TQDM",
        "tqdm",
        "tqdmn",
        "ObjDict",
        "LogFile",
    ):
        assert hasattr(imports, obj), f"{obj}"

    return


# /def


def test_import_base_has_ipython():
    """Test base imports when in ipython environment.

    combines with test_import_base to test all imports.

    """
    from utilipy.imports import base_imports as imports

    try:
        get_ipython()
        if get_ipython() is None:  # double checking
            raise NameError

    except NameError:
        pass

    else:
        for obj in (
            "InteractiveShell",
            "set_trace",
            "display",
            "Latex",
            "Markdown",
            "printmd",
            "printMD",
            "printltx",
            "printLaTeX",
            "set_autoreload",
            "aimport",
            "run_imports",
            "import_from_file",
            "add_raw_code_toggle",
        ):
            assert hasattr(imports, obj)

    return


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for :mod:`~utilipy.utils`."""


__all__ = [
    "test_init_data_utils",
    "test_init_decorators",
    "test_init_imports",
    "test_init_ipython",
    "test_init_math",
    "test_init_plot",
    "test_init_scripts",
    "test_init_utils",
    "test_utils_top_level_imports",
]


##############################################################################
# IMPORTS

# BUILT-IN
import os

# PROJECT-SPECIFIC
from utilipy import (
    data_utils,
    decorators,
    imports,
    ipython,
    math,
    plot,
    scripts,
    utils,
)

##############################################################################
# TESTS
##############################################################################


def test_init_data_utils():
    """Test :mod:`~utilipy.data_utils` initialization."""
    # Expectations
    local = [
        # modules
        "crossmatch",
        "decorators",
        "select",
        "fitting",
        "utils",
        "xfm",
        # decorators
        "idxDecorator",
        # data transformation graph
        "data_graph",
        "TransformGraph",
        "DataTransform",
        # xmatch
        "indices_xmatch_fields",
        "xmatch_fields",
        "xmatch",
        "non_xmatched",
        # utils
        "get_path_to_file",
        "make_shuffler",
    ]
    local += data_utils.select.__all__

    # test __all__ conforms to module
    for name in data_utils.__all__:
        assert hasattr(data_utils, name)

    # test __all__ matches expectations
    for name in data_utils.__all__:
        assert name in local

    return


# /def


# --------------------------------------------------------------------------


def test_init_decorators():
    """Test :mod:`~utilipy.decorators` initialization."""
    # Expectations
    local = [
        # modules
        "baseclass",
        "docstring",
        "func_io",
        "code_dev",
        # defined here
        "store_function_input",
        "add_folder_backslash",
        # baseclass
        "DecoratorBaseClass",
        "classy_decorator",
        # data-type decorators
        "dtypeDecorator",
        "dtypeDecoratorMaker",
        "intDecorator",
        "floatDecorator",
        "strDecorator",
        "boolDecorator",
        "ndarrayDecorator",
        "ndfloat64Decorator",
        # convenience
        "functools",
        "inspect",
        "wraps",
    ]

    # test __all__ conforms to module
    for name in decorators.__all__:
        assert hasattr(decorators, name)

    # test __all__ matches expectations
    for name in decorators.__all__:
        assert name in local

    return


# /def


# --------------------------------------------------------------------------


def test_init_imports():
    """Test :mod:`~utilipy.imports` initialization."""
    # Expectations
    local = [
        "conf",
        "use_import_verbosity",
    ]

    # test __all__ conforms to module
    for name in imports.__all__:
        assert hasattr(imports, name)

    # test __all__ matches expectations
    for name in imports.__all__:
        assert name in local

    return


# /def


# --------------------------------------------------------------------------


def test_init_ipython():
    """Test :mod:`~utilipy.ipython` initialization."""
    # Expectations
    local = [
        "ipython_help",
        "get_ipython",
        "InteractiveShell",
        "set_trace",
        "display",
        "Latex",
        "Markdown",
        "HTML",
        "set_autoreload",
        "aimport",
        "run_imports",
        "import_from_file",
        "add_raw_code_toggle",
        "printMD",
        "printLTX",
    ]

    # test __all__ conforms to module
    for name in ipython.__all__:
        assert hasattr(ipython, name), name

    # test __all__ matches expectations
    for name in ipython.__all__:
        assert name in local

    return


# /def


# --------------------------------------------------------------------------


def test_init_math():
    """Test :mod:`~utilipy.math` initialization."""
    # Expectations
    local = []
    local += math.core.__all__

    # test __all__ conforms to module
    for name in math.__all__:
        assert hasattr(math, name)

    # test __all__ matches expectations
    for name in math.__all__:
        assert name in local

    return


# /def


# --------------------------------------------------------------------------


def test_init_plot():
    """Test :mod:`~utilipy.plot` initialization."""
    # Expectations
    local = []

    # test __all__ conforms to module
    for name in plot.__all__:
        assert hasattr(plot, name)

    # test __all__ matches expectations
    for name in plot.__all__:
        assert name in local

    return


# /def


# --------------------------------------------------------------------------


def test_init_scripts():
    """Test :mod:`~utilipy.scripts` initialization."""
    # Expectations
    local = []

    # test __all__ conforms to module
    for name in scripts.__all__:
        assert hasattr(scripts, name)

    # test __all__ matches expectations
    for name in scripts.__all__:
        assert name in local

    return


# /def


# --------------------------------------------------------------------------


def test_init_utils():
    """Test :mod:`~utilipy.utils` initialization."""
    # Expectations
    local = [
        # modules
        "collections",
        "logging",
        "exceptions",
        "functools",
        "inspect",
        "metaclasses",
        "misc",
        "pickle",
        "string",
        "typing",
        # classes and functions
        "LogPrint",
        "LogFile",
        "ObjDict",
        "WithDocstring",
        "WithMeta",
        "WithReference",
        "temporary_namespace",
        "make_help_function",
    ]

    # test __all__ conforms to module
    for name in utils.__all__:
        assert hasattr(utils, name)

    # test __all__ matches expectations
    for name in utils.__all__:
        assert name in local

    return


# /def


def test_utils_top_level_imports():
    """Test Top-Level Imports."""
    # First test they exist
    subpkg: str
    for subpkg in utils.__all_top_imports__:
        assert hasattr(utils, subpkg)

    # Next test that top-levels are all the possible top-levels
    drct: str = os.path.split(utils.__file__)[0]  # directory
    donottest = ("tests", "__pycache__")  # stuff not to test

    for file in os.listdir(drct):  # iterate through directory
        # test?
        if os.path.isdir(drct + "/" + file) and file not in donottest:
            assert file in utils.__all_top_imports__
        # else:  # nope, chuck testa.
        #     pass

    return


# /def


# --------------------------------------------------------------------------


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : ipython.notebook
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

"""Functions for Jupyter notebook / lab / hub.

Routine Listings
----------------
`add_raw_code_toggle`
    Add a toggle to show/hide code cells when Notebook is exported to HTML

References
----------
IPython [#]_, Jupyter Notebook [#]_

.. [#] Fernando Pérez, Brian E. Granger, IPython: A System for Interactive
    Scientific Computing, Computing in Science and Engineering, vol. 9,
    no. 3, pp. 21-29, May/June 2007, doi:10.1109/MCSE.2007.53.
    URL: https://ipython.org
.. [#] Jupyter Notebook

"""

__credits__ = ["Jo Bovy"]

__all__ = [
    "add_raw_code_toggle",
]


##############################################################################
# IMPORTS

# THIRD PARTY
from IPython.display import HTML

##############################################################################
# CODE


def add_raw_code_toggle() -> HTML:
    """Add a toggle for code cells when Notebook is exported to HTML.

    HTML and Javascript button, placed directly below function execution cell.

    Returns
    -------
    HTML

    References
    ----------
    code from Jo Bovy

    """
    return HTML(
        """<script>
    code_show=true;
    function code_toggle() {
     if (code_show){
     $('div.input').hide();
     } else {
     $('div.input').show();
     }
     code_show = !code_show
    }
    $( document ).ready(code_toggle);
    </script>
    <form action="javascript:code_toggle()">
    <input type="submit" value="Click here to toggle on/off the raw code.">
    </form>"""
    )


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Typing library utilities."""

__all__ = [
    # compat
    "OrderedDictType",
    "EllipsisType",
    # numpy
    "array_like",
    # Astropy
    "CoordinateRepresentationType",
    "CoordinateDifferentialType",
    "CoordinateType",
    "FrameOptionsType",
    "TableType",
    "UnitableType",
    "DistanceType",
    "AngleType",
    "ModelType",
    "FittableModelType",
]


###############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from typing import *  # noqa

# THIRD PARTY
import astropy.coordinates as coord
import astropy.units as u
import numpy as np
from astropy import table
from astropy.modeling.core import CompoundModel, FittableModel, Model

###############################################################################
# CODE
###############################################################################

OrderedDictType = T.OrderedDict
EllipsisType = type(Ellipsis)


#####################################################################
# Numpy


array_like = T.TypeVar("array_like", np.ndarray, list, tuple, T.Sequence)
"""Array-like types compatible with [numpy]_'s "array-like".

References
----------
.. [numpy] Travis E, Oliphant. A guide to NumPy: Trelgol Publishing, (2006).

Examples
--------

    >>> sc: FrameOptionsType = SkyCoord()
    >>> isinstance(sc, SkyCoord)
    True

"""


#####################################################################
# Astropy

CoordinateRepresentationType = T.TypeVar(
    "RepresentationType",
    coord.BaseRepresentation,
    coord.BaseRepresentationOrDifferential,
)
"""Astropy [astropy]_ representation types.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

"""


CoordinateDifferentialType = T.TypeVar(
    "RepresentationType",
    coord.BaseDifferential,
    coord.BaseRepresentationOrDifferential,
)
"""Astropy [astropy]_ representation types.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

"""


# -------------------------------------------------------------------

CoordinateType = T.TypeVar(
    "CoordinateType", coord.BaseCoordinateFrame, coord.SkyCoord
)
"""Astropy [astropy]_ coordinate frame-like types.

Subclasses of BaseCoordinateFrame or SkyCoord.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> sc: CoordinateType = SkyCoord()
    >>> isinstance(sc, SkyCoord)
    True

"""

FrameOptionsType = T.TypeVar("FrameOptionsType", str, CoordinateType)
"""Astropy [astropy]_ frame-like types.

Subclasses of BaseCoordinateFrame or SkyCoord, or a string.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> sc: FrameOptionsType = SkyCoord()
    >>> isinstance(sc, SkyCoord)
    True

"""

# -------------------------------------------------------------------

TableType = T.TypeVar("TableType", table.Table, table.QTable)
"""Astropy [astropy]_ Table-types.

Subclasses of Table, QTable.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> tbl: TableType = QTable()
    >>> isinstance(tbl, Table)
    True

"""

# -------------------------------------------------------------------

UnitableType = T.TypeVar("UnitableType", u.Unit, u.Quantity, str)
"""Types that work with Astropy [astropy]_ :class:`~astropy.units.Unit`.

Subclasses of Unit, Quantity.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> x: UnitableType = 10 * u.km
    >>> isinstance(x, u.Quantity)
    True

"""

UnitType = T.TypeVar("UnitType", u.UnitBase, u.Unit)
""":class:`~astropy.units.Unit` Type [astropy]_.

Subclasses of Unit.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> unit: UnitType = u.km
    >>> isinstance(x, u.Unit)
    True

"""

QuantityType = T.TypeVar("QuantityType", u.Quantity, u.SpecificTypeQuantity)
"""Types for :class:`~astropy.units.Quantity`[astropy]_ .

Subclasses of Distance, Quantity.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> x: QuantityType = 10 * u.km
    >>> isinstance(x, u.Quantity)
    True

"""

DistanceType = T.TypeVar("DistanceType", coord.Distance, u.Quantity)
"""Types for :class:`~astropy.coordinates.Distance`[astropy]_ .

Subclasses of Distance, Quantity.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> x: DistanceType = 10 * u.km
    >>> isinstance(x, u.Quantity)
    True

"""


AngleType = T.TypeVar("DistanceType", coord.Angle, u.Quantity)
"""Types for :class:`~astropy.coordinates.Angle`[astropy]_ .

Subclasses of Angle, Quantity.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> x: DistanceType = 10 * u.deg
    >>> isinstance(x, u.Quantity)
    True

"""

# -------------------------------------------------------------------

ModelType = T.TypeVar("ModelType", Model, CompoundModel)
"""Astropy :class:`~astropy.modeling.Model` Types [astropy]_.

Subclasses of Model, CompoundModel.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> from astropy.modeling.models import Model, Gaussian1D
    >>> x: ModelType = Gaussian1D()
    >>> isinstance(x, Model)
    True

"""

FittableModelType = T.TypeVar("ModelType", FittableModel, CompoundModel)
"""Astropy :class:`~astropy.modeling.FittableModel` Types [astropy]_.

Subclasses of FittableModel, CompoundModel.
Note that not all CompoundModels are fittable, but this can't be checked.

References
----------
.. [astropy] Astropy Collaboration et al., 2018, AJ, 156, 123.

Examples
--------

    >>> from astropy.modeling.models import Model, Gaussian1D
    >>> x: ModelType = Gaussian1D()
    >>> isinstance(x, Model)
    True

"""


###############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : base_imports
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

# Docstring and Metadata
"""Base set of imports.

Returns
-------
Base: imports

    - os, sys, time, pdb, warnings,
    - numpy -> np, scipy,
    - tqdm -> TQDM, tqdm, .tqdm_notebook -> tqdmn

IPython: imports

    - display, Latex, Markdown, set_trace,
    - printmd, printMD, printltx, printLaTeX,
    - set_autoreload, aimport,
    - run_imports, import_from_file,
    - add_raw_code_toggle

utilipy: imports

    - LogFile
    - ObjDict

References
----------
SciPy references are [#]_ and [#]_.

NumPy references are [#]_ and [#]_.

IPython reference is [#]_.

Matplotlib reference is [#]_.

.. [#] Travis E. Oliphant, "Python for Scientific Computing", Computing in
   Science & Engineering, 9, 10-20 (2007), DOI:10.1109/MCSE.2007.58
   http://scitation.aip.org/content/aip/journal/cise/9/3/10.1109/MCSE.2007.58.

.. [#] K. Jarrod Millman and Michael Aivazis, "Python for Scientists and
   Engineers", Computing in Science & Engineering, 13, 9-12 (2011),
   DOI:10.1109/MCSE.2011.36
   http://scitation.aip.org/content/aip/journal/cise/13/2/10.1109/MCSE.2011.36.

.. [#] Travis E, Oliphant. A guide to NumPy, USA: Trelgol Publishing, (2006).

.. [#] Stéfan van der Walt, S. Chris Colbert and Gaël Varoquaux.
   "The NumPy Array: A Structure for Efficient Numerical Computation",
   Computing in Science & Engineering, 13, 22-30 (2011),
   DOI:10.1109/MCSE.2011.37
   http://scitation.aip.org/content/aip/journal/cise/13/2/10.1109/MCSE.2011.37.

.. [#] Fernando Pérez, Brian E. Granger, "IPython: A System for Interactive
   Scientific Computing", Computing in Science and Engineering, vol. 9,
   no. 3, pp. 21-29, May/June 2007, doi:10.1109/MCSE.2007.53.
   URL: https://ipython.org.

.. [#] John D. Hunter, Matplotlib: A 2D Graphics Environment, Computing in
   Science & Engineering, 9, 90-95 (2007), DOI:10.1109/MCSE.2007.55
   http://scitation.aip.org/content/aip/journal/cise/9/3/10.1109/MCSE.2007.55.

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    # modules
    "os",
    "sys",
    "time",
    "warnings",
    "np",
    "scipy",
    "TQDM",
    "tqdm",
    "tqdmn",
    "conf",
    "ObjDict",
    "make_help_function",
    "LogFile",
    # functions
    "base_imports_help",
]


##############################################################################
# IMPORTS

# BUILT-IN
import os
import sys  # operating system
import time  # timing
import warnings  # warning

# THIRD PARTY
import numpy as np  # numerical python
import scipy  # scientific python
import tqdm as TQDM
from tqdm import tqdm as tqdm
from tqdm import tqdm_notebook as tqdmn

# PROJECT-SPECIFIC
from utilipy.imports import conf
from utilipy.utils import ObjDict  # custom dictionary-like object
from utilipy.utils import make_help_function
from utilipy.utils.logging import LogFile  # LoggerFile  # custom logging

# from tqdm.autonotebook import tqdm


##############################################################################

# +--------------------------------------------------------------------------+
# IPython

try:

    HAS_IPYTHON: bool = False

    get_ipython()

    if get_ipython() is None:  # double checking
        raise NameError

except NameError:

    pass

else:

    HAS_IPYTHON = True

    from IPython.core.interactiveshell import InteractiveShell
    from IPython.core.debugger import set_trace
    from IPython.display import (
        display,  # display is a better print
        Latex,
        Markdown,  # for printing LaTeX or Markdown strings
    )

    # %run runs in the main namespace, so need to run as 'src.', not '.''
    from utilipy.ipython import (
        printMD,  # markdown printing
        printLTX,  # LaTeX printing
        set_autoreload,
        aimport,  # imports
        run_imports,
        import_from_file,  # imports
        add_raw_code_toggle,  # notebook
    )

    __all__ += [
        "InteractiveShell",
        "set_trace",
        "display",
        "Latex",
        "Markdown",
        "printMD",
        "printLTX",
        "set_autoreload",
        "aimport",
        "run_imports",
        "import_from_file",
        "add_raw_code_toggle",
    ]


##############################################################################
# Running Imported Functions

if HAS_IPYTHON:

    InteractiveShell.ast_node_interactivity = "all"


##############################################################################
# Printing Information

base_imports_help = make_help_function("base", __doc__, look_for="Returns")


if conf.verbose_imports:
    base_imports_help()

##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : IPython printing
#
# ----------------------------------------------------------------------------

# Docstring and Metadata
"""functions for enhanced printing in an IPython environment.

Routine Listings
----------------
printMD
    Print in Markdown.
printLTX, printLaTeX
    Print in LaTeX.

References
----------
Ipython [1]_

.. [1] Fernando Pérez, Brian E. Granger, IPython: A System for Interactive
    Scientific Computing, Computing in Science and Engineering, vol. 9,
    no. 3, pp. 21-29, May/June 2007, doi:10.1109/MCSE.2007.53.
    URL: https://ipython.org

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "printMD",
    "printLTX",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T

# THIRD PARTY
from IPython.display import Latex, Markdown

from IPython.display import display  # , HTML; display is a better print

##############################################################################
# CODE
##############################################################################


def printMD(
    s: str,
    color: T.Optional[str] = None,
    size: T.Optional[float] = None,
    bold: bool = False,
    italic: bool = False,
    fontweight: T.Optional[float] = None,
    fontstyle: T.Optional[str] = None,
    highlight: T.Optional[str] = None,
):
    """Print in Markdown.

    uses <span>

    Parameters
    ----------
    s : str
        the string to print
    color :  str or None, optional  (default None)
        sets the 'style:color'
        for color names see https://www.w3schools.com/tags/ref_colornames.asp
        or can supply hex value
    size : int, optional  (default None)
        sets the 'style:font-size' in px
    bold : bool, optional
        (default False)
        shortcut to fontweight='bold'
        *fontweight* takes precedence
    italic : bool, optional
        (default False)
        shortcut to fontstyle='italic'
        *fontstyle* takes precedence
    fontweight : str or int, optional
        (default None)
        sets the 'style:font-weight'
        str options: 'normal', 'bold', 'lighter', 'bolder'
        int options: 0-1000
    fontstyle : str or None, optional
        (default None)
        sets the 'style:font-style'
        see https://www.w3schools.com/cssref/pr_font_font-style.asp
        str options: normal, italic, oblique, initial, inherit
    highlight : str or None, optional
        (default None)

    """
    # doing style
    styleattrs = []
    if color is not None:
        styleattrs.append(f"color:{color}")
    if size is not None:
        styleattrs.append(f"font-size:{size}px")
    # if textalign is not None:  # FIXME doesn't work in jupyter-lab
    #    styleattrs.append(f'text-align:{textalign}')

    if fontweight is not None:
        styleattrs.append(f"font-weight:{fontweight}")
    elif bold:
        styleattrs.append("font-weight:bold")

    if fontstyle is not None:
        styleattrs.append(f"font-style:{fontstyle}")
    elif italic:
        styleattrs.append("font-style:italic")

    if highlight is not None:
        styleattrs.append("background-color:{highlight}")

    # TODO: add more options

    style = "style='{}'".format(";".join(styleattrs))

    string = f"<span {style}>{s}</span>"
    display(Markdown(string))


# /def


# printHTML = printmd


# ----------------------------------------------------------------------------


def printLTX(
    s: str,
    math: T.Union[str, bool] = False,
    equation: T.Union[str, bool] = False,
    matrix: T.Union[str, bool] = False,
    label: str = "",
):
    r"""Print in LaTeX.

    Parameters
    ----------
    s : raw! str
        the s to print
    math : bool, str  (default False)
        whether the whole string should be wrapped in $$
        if str: wrap in that string
    equation: bool, None  (default False)
        whether the whole string should be wrapped in $\equation$
        if None: then '\equation*'
    matrix: bool, str  (default False)
        whether the whole string should be wrapped in $\matrix$
        if str, use that matrix type, like 'bmatrix'
        shortcuts) 'b' -> 'bmatrix'
    label: the label of the equation, etc.
        only used in equation or matrix

    Notes
    -----
    .. todo::

        optional PyLaTeX input

    """
    if label == "":
        pass
    else:
        label = r"\label{" + label + "}"

    if isinstance(math, str):
        s = math + s + math
    elif math:  # (True)
        s = "$" + s + "$"

    if equation:
        s = r"\begin{equation}" + label + s + r"\end{equation}"
    elif equation is None:
        s = r"\begin{equation*}" + label + s + r"\end{equation*}"

    if isinstance(matrix, str):
        if matrix == "b":
            matrix = "bmatrix"
        s = r"\begin{" + matrix + "}" + label + s + r"\end{" + matrix + "}"
    elif matrix:
        s = r"\begin{matrix}" + label + s + r"\end{matrix}"

    display(Latex(s))

    return


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Docstring decorators."""


__all__ = [
    # "format_doc",
    "set_docstring_for_import_func",
]


##############################################################################
# IMPORTS

# BUILT-IN
import ast
import typing as T

# PROJECT-SPECIFIC
from utilipy.data_utils import get_path_to_file
from utilipy.utils import functools

##############################################################################
# CODE
##############################################################################


def set_docstring_for_import_func(
    *path: str, package: T.Optional[str] = None, section: str = "Returns"
) -> str:
    """Set docstring for IPython import function, from import file's docstring.

    Takes a helper function for a module and adds the content of the modules'
    `section`. This currently only works on numpy-doc style docstrings.

    Parameters
    ----------
    *path: str
        path of import module
    package : str, optional, keyword only
        package for :func:`~utilipy.data_utils.get_path_to_file`
    section: str, optiona, keyword only
        numpy-doc style section name

    Notes
    -----
    This function might be moved

    """
    module = get_path_to_file(*path, package=package)

    # read docstring out of file
    with open(module, "r") as fd:
        module_doc = ast.get_docstring(ast.parse(fd.read()))

    # process docstring
    len_title = 2 * len(section) + 1  # section & underline
    ind_section = module_doc.find(section)
    if ind_section == -1:
        raise IndexError(f"Section {section} does not exist.")

    ind = ind_section + len_title
    end_ind = ind + module_doc[ind:].find("---")  # noqa

    sub_doc = module_doc[ind:end_ind]  # get section (+ next header)

    # drop next header
    sec = "\n".join(sub_doc.split("\n")[:-2])

    # modify function with a basic decorator
    def decorator(func):
        @functools.wraps(func, docstring=(func.__doc__ or "") + "\n" + sec)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    # /def

    return decorator


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""`data_util` decorators."""


##############################################################################
# IMPORTS

# BUILT-IN
import sys
import typing as T

# THIRD PARTY
import numpy as np
from typing_extensions import Literal

# PROJECT-SPECIFIC
from utilipy.utils import functools

##############################################################################
# PARAMETERS

# Windows prints numpy differently. Numpy outputs need to be ignored.
# TODO: a better platform-specific fix for numpy.
if sys.platform.startswith("win"):
    __doctest_skip__ = ["idxDecorator"]


##############################################################################
# CODE
##############################################################################


def idxDecorator(
    function: T.Optional[T.Callable] = None,
    *,
    as_ind: T.Union[bool, Literal["flatten"]] = False,
    _doc_fmt: T.Optional[dict] = None,  # ibid
    _doc_style: T.Union[str, T.Callable, None] = None,  # ibid
) -> T.Callable:
    """Control whether to return boolean array or indices.

    for functions which return bool arrays
    adds `as_ind` as a kwarg to decorated function

    Parameters
    ----------
    function : Callable, optional
        (default None)
        the function to be decoratored
        if None, then returns decorator to apply.
    as_ind : bool or "flatten", optional
        (default False)
        whether to return bool array or indices
        (``where(bool array == np.True_)``)
        if "flatten", flattens a nested list with only 1 element
        ie ([0], ) -> [0]
        sets the default behavior for the wrapped `function`

    Returns
    -------
    wrapper : Callable
        wrapper for `function`
        includes the original function in a method ``.__wrapped__``

    Other Parameters
    ----------------
    _doc_fmt : dict, optional
        docstring formatter
        argument into :func:`~utilipy.utils.functools.wraps`
    _doc_style : str or Callable, optional
        docstring style
        argument into :func:`~utilipy.utils.functools.wraps`

    Notes
    -----
    Adds `as_ind` and other parameters to the function signature and docstring.

    Examples
    --------
    Use the Standard Decorator:

        >>> x = np.array([0, 2])
        >>> @idxDecorator
        ... def func1(x):
        ...     return x < 1

    calling normally
        >>> func1(x) # doctest: +SKIP
        array([ True, False], dtype=bool))

    using added kwarg
        >>> func1(x, as_ind=True)
        (array([0]),)

    and flattening
        >>> func1(x, as_ind="flatten")
        array([0])

    Set a Different Default:

        >>> @idxDecorator(as_ind=True)
        ... def func2(x):
        ...     return x < 1

        >>> func2(x)
        (array([0]),)

        >>> func2(x, as_ind=False) # doctest: +SKIP
        array([ True, False])

    Making a New Decorator:

        >>> trueidxdec = idxDecorator(as_ind="flatten")
        >>> @trueidxdec
        ... def func3(x):
        ...     return x < 1
        >>> func3(x)
        array([0])
        >>> func3(x, as_ind=False) # doctest: +SKIP
        array([ True, False])

    Wrapping Existing Functions

        >>> def func(x):
        ...     return x < 1
        >>> newfunc = idxDecorator(func, as_ind=True)
        >>> newfunc(x)
        (array([0]),)
        >>> newfunc(x, as_ind=False) # doctest: +SKIP
        array([ True, False])

    """
    if function is None:  # allowing for optional arguments
        return functools.partial(
            idxDecorator,
            as_ind=as_ind,
            _doc_fmt=_doc_fmt,
            _doc_style=_doc_style,
        )

    @functools.wraps(function, _doc_fmt=_doc_fmt, _doc_style=_doc_style)
    def wrapper(
        *args: T.Any, as_ind: bool = as_ind, **kwargs: T.Any
    ) -> T.Sequence:
        """Index-decorator wrapper docstring, overwritten by `function`.

        Other Parameters
        ----------------
        as_ind : bool or "flatten", optional
            (default {as_ind})
            whether to return a boolean array, or array of indices.

        Raises
        ------
        `~ValueError`
            if `as_ind` is "flatten" and cannot be flattened (len > 1)

        """
        # Step 1: call function
        bool_arr = function(*args, **kwargs)

        # Step 2:  determine whether to return indices or bool array
        if as_ind:  # works for bool or full str
            # get the indices
            inds = np.nonzero(np.asarray(bool_arr) == np.True_)

            # determine whether to return as-is, or flatten
            if as_ind == "flatten":
                if len(inds) == 1:  # nested list with only 1 element
                    return inds[0]
                else:  # not a valid option
                    raise ValueError
            else:  # do not flatten
                return inds

        # return a bool array
        return bool_arr

    # /def

    return wrapper


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-
# see LICENSE.rst

"""Utilipy Utils Tests."""


__all__ = [
    "test_typing",
    "test_string",
    "test_pickle",
    "test_inspect",
    "test_functools",
    "test_exceptions",
    "test_collections",
    "_test_functools_util",
]


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from . import (
    _test_functools_util,
    test_collections,
    test_exceptions,
    test_functools,
    test_inspect,
    test_pickle,
    test_string,
    test_typing,
)

##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : logfile_print
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

# Docstring and Metadata
"""LogPrint function for logging to the console."""


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T

##############################################################################
# LogPrint


class LogPrint(object):
    """A basic logger wrapper for print."""

    def __init__(
        self,
        verbose: int = 0,
        sec_div: str = "-",
        header: T.Union[str, bool, None] = None,
        show_header: bool = True,
        # for compatibility
        **kw: T.Any,
    ) -> None:
        """Initialize LogPrint.

        start the file header (just printing)

        """
        super().__init__()

        # keeping input arguments
        self.verbose = verbose
        self.sec_div = sec_div

        # making file header
        if header is not False:
            if header in {None, True}:
                header = ""
            elif isinstance(header, str):
                if header[-1] != " ":  # making sure ends in space
                    header += " "
            else:
                raise TypeError

            self.write(f"{header}Log:", endsection="=", print=show_header)

    # /def

    # ------------------------------------------------------------------------

    @classmethod
    def open(
        cls,
        verbose: int = 0,
        sec_div: str = "-",
        header: T.Union[bool, None, str] = None,
        show_header: bool = True,
        **kw: T.Any,
    ) -> T.Any:
        """Open LogPrint.

        .. todo::

            kw absorbs all extra kwargs to be consistent with LogFile

        """
        return cls(
            verbose=verbose,
            sec_div=sec_div,
            header=header,
            show_header=show_header,
        )

    # /def

    @classmethod
    def open_to_write(
        cls,
        verbose: int = 0,
        sec_div: str = "-",
        header: T.Union[bool, None, str] = None,
        show_header: bool = True,
        **kw: T.Any,
    ) -> T.Any:
        """open_to_write.

        .. todo::

            kw absorbs all extra kwargs to be consistent with LogFile

        """
        return cls(
            verbose=verbose,
            sec_div=sec_div,
            header=header,
            show_header=show_header,
        )

    # /def

    @classmethod
    def open_to_read(cls) -> T.Any:
        """Open a logfile to read.

        this class uses ``open``, not a more extensive logger, like ``logging``

        The arguments filename - opener are all for ``open``

        """
        return cls()

    # /def

    # ------------------------------------------------------------------------

    def print(
        self, *text: str, start: str = "", sep: str = " ", end: str = "\n"
    ) -> None:
        r"""Wrap-function for print.

        Parameters
        ----------
        text: str
            the text to print
        sep: str  (default ' ')
            the separater for print
        end: str  (default '\n')
            the end for print

        """
        print(start, end="")
        print(*text, sep=sep, end=end)

    # /def

    def _write(self, *string, start="", sep=" ", end="\n") -> None:
        """Writer method.

        this is implemented solely to be overwritten by child classes

        """

    # /def

    def _print_and_write(self, *string, start="", sep=" ", end="\n") -> None:
        """Help method to print and call _write.

        this is implemented solely to be overwritten by child classes

        """
        self.print(*string, start=start, sep=sep, end=end)  # printing
        self._write(*string, start=start, sep=sep, end=end)  # writing

    # /def

    def newsection(self, title=None, div=None, print=True) -> None:
        """Make new section.

        Parameters
        ----------
        div: str, None  (default None)
            the divider
            None -> sec_div from initialization

        """
        if title is None:
            title = ""

        if div is None:  # get default divider if custom one not provided
            div = self.sec_div

        full_div = div * int(79 / len(div))  # round out to full line length

        if print:
            self._print_and_write(
                full_div, title, start="\n", sep="\n", end="\n"
            )
        else:
            self._write(full_div, title, start="\n", sep="\n", end="\n\n")

    # /def

    def write(
        self,
        *text: str,
        start: str = "",
        sep: str = " ",
        end: str = "\n",
        startsection: T.Union[bool, str] = False,
        endsection: T.Union[bool, str] = False,
        print: bool = True,
    ) -> None:
        r"""Print text.

        Parameters
        ----------
        text: str
            the text to write & print
        start: str  (default '')
            start to print
        sep: str  (default ' ')
            the separater for print
        end: str  (default '\n')
            the end for print
        startsection: bool  (default False)
            whether to start a new section before writing
        endsection:  bool  (default False)

        """
        if startsection is not False:
            if isinstance(startsection, str):
                self.newsection(div=startsection, print=print)
            else:
                self.newsection(print=print)

        if print:
            self._print_and_write(*text, start=start, sep=sep, end=end)
        else:
            self._write(*text, start=start, sep=sep, end=end)

        if endsection is not False:
            if isinstance(endsection, str):
                self.newsection(div=endsection, print=print)
            else:
                self.newsection(print=print)

    # /def

    def record(
        self,
        *text: str,
        start: str = "",
        sep: str = " ",
        end: str = "\n",
        startsection: T.Union[bool, str] = False,
        endsection: T.Union[bool, str] = False,
    ) -> None:
        """Redirects to write, which goes to print.

        this is implemented solely for compatibility with LogFile

        """
        self.write(
            *text,
            start=start,
            sep=sep,
            end=end,
            startsection=startsection,
            endsection=endsection,
            print=False,
        )

    # /def

    # ------------------------------------------------------------------------

    def report(
        self,
        *msgs: str,
        verbose: T.Optional[int] = None,
        print: bool = True,
        write: bool = True,
        start_at: int = 1,
        **kw: T.Any,
    ) -> None:
        """Report function whose message is determined by `verbose`.

        Parameters
        ----------
        msgs : str
            the verbosity-ordered messages
            blank messages can be <None>, not only ''
        verbose : int, optional
            which message to record
            None (default) uses self.verbose (default = 0, unless specified)
        print : bool
            whether to print, or just record
        write : bool
            whether to write to logger file
            write redirects to print in this class
        start_at : int
            what level of verbosity is the first `msg`
            ex: report('test', start_at=3) means 'test' is at verbose=3
        kw:
            kwargs for self.write or self.print

        """
        if verbose is None:
            verbose = self.verbose

        if verbose < start_at:  # below starting verbosity
            msg = None
        elif verbose - start_at >= len(msgs):  # above / at last message
            msg = msgs[-1]
        else:  # inside message options
            msg = msgs[verbose - start_at]

        if str(msg) == "":
            msg = None  # catching null messages

        if msg is not None:
            if write:
                self.write(msg, print=print, **kw)
            else:
                self.print(msg, **kw)

    # /def

    # ------------------------------------------------------------------------

    def close(self):
        """Close the non-existent file.

        implemented solely to be overwritten by child classes.

        """

    # /def

    @classmethod
    def read_log(
        cls,
        filename: str,
        buffering: int = -1,
        encoding: T.Any = None,
        errors: T.Any = None,
        newline: T.Any = None,
        closefd: bool = True,
        opener: None = None,
    ) -> str:
        """Read and print out a previous log.

        Parameters
        ----------
        filename: str
            the file name / path at which to save this log

        .. todo::

            what should this return?

        """
        file = open(
            filename,
            mode="r",
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
        )
        log = file.read()

        print(log)
        return log

    # /def


# /class

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Test data_util :mod:`~utilipy.data_utils.decorators`."""


__all__ = [
    "test_idxDecorator_standard",
    "test_idxDecorator_defaults",
    "test_idxDecorator_new_decorator",
    "test_idxDecorator_existing_function",
]


##############################################################################
# IMPORTS

# THIRD PARTY
import numpy as np

# PROJECT-SPECIFIC
from utilipy.data_utils.decorators import idxDecorator

##############################################################################
# PARAMETERS

x = np.arange(2)
y = np.arange(2) + 2

z = np.c_[x, y].T


##############################################################################
# TESTS
##############################################################################

##########################################################################
# idxDecorator


def test_idxDecorator_standard():
    """Test standard use of idxDecorator."""
    # defining function
    @idxDecorator
    def func1(x):
        return x < 1

    # /def

    # calling normally
    assert all(func1(x) == np.array([True, False]))

    # using added kwarg
    assert all(func1(x, as_ind=True) == np.array([0]))

    return


# /def


# ------------------------------------------------------------------------


def test_idxDecorator_defaults():
    """Test setting default in idxDecorator."""
    # defining function
    @idxDecorator(as_ind=True)
    def func2(x):
        return x < 1

    # /def

    # calling normally, defaulted to index
    assert all(func2(x) == np.array([0]))

    # using added kwarg
    assert all(func2(x, as_ind=False) == np.array([True, False]))

    return


# /def


# ------------------------------------------------------------------------


def test_idxDecorator_new_decorator():
    """Test making new decorator."""
    # making new decorator with different value
    trueidxdec = idxDecorator(as_ind=True)

    # defining function
    @trueidxdec
    def func3(x, **kw):  # kw only for Codacy code quality check
        return x < 1

    # /def

    # calling normally, defaulted to index
    assert func3(x) == np.array([0])

    # using added kwarg
    assert all(func3(x, as_ind=False) == np.array([True, False]))

    return


# /def


# ------------------------------------------------------------------------


def test_idxDecorator_existing_function():
    """Test wrapping existing function with idxDecorator."""
    # defining function
    def func(x):
        return x < 1

    # /def

    # wrapping existing function
    newfunc = idxDecorator(func, as_ind=True)

    # calling normally, defaulted to index
    assert newfunc(x) == np.array([0])

    # using added kwarg
    assert all(newfunc(x, as_ind=False) == np.array([True, False]))

    return


# /def

# ------------------------------------------------------------------------


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : ipython.plot
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

# Docstring and Metadata
"""Plotting in an IPython environment.

Routine Listings
----------------
configure_matplotlib
    configure matplotlib Jupyter magic setting.

References
----------
Ipython [1]_

.. [1] Fernando Pérez, Brian E. Granger, IPython: A System for Interactive
    Scientific Computing, Computing in Science and Engineering, vol. 9,
    no. 3, pp. 21-29, May/June 2007, doi:10.1109/MCSE.2007.53.
    URL: https://ipython.org

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "configure_matplotlib",
]


##############################################################################
# IMPORTS

# BUILT-IN
import warnings

# THIRD PARTY
from IPython import get_ipython

##############################################################################
# CODE


def configure_matplotlib(
    backend: str = "inline", figure_format: str = "retina"
) -> None:
    """Configure Matplotlib Jupyter magic setting.

    Parameters
    ----------
    backend : str, optional
        (defualt 'inline')
        set the matplotlib backend

    """
    if get_ipython() is None:
        warnings.warn(
            "not in an IPython environment," "cannot configure matplotlib"
        )
        return

    from IPython.terminal.pt_inputhooks import UnknownBackend

    try:
        # set matplotlib backend
        get_ipython().magic(f"matplotlib {backend}")

    except UnknownBackend:
        warnings.warn("not a valid backed. Please try again.")

    else:
        # set the resolution
        get_ipython().magic(
            "config InlineBackend.figure_format" f"='{figure_format}'"
        )

    return


# /def

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Baseclass for tests which rely on a class."""

__all__ = [
    "BaseClassDependentTests",
]


##############################################################################
# CODE
##############################################################################


class BaseClassDependentTests:
    """Base class for tests which rely on a class.

    Subclasses must specify a class in the class definition.

        >>> class SubClass(BaseClassDependentTests, klass=object):
        ...     '''A SubClass with class=object.'''
        ...     pass
        >>> SubClass.klass == object
        True

    """

    def __init_subclass__(cls, klass: object, **kwargs):
        """Initialize subclass.

        Parameters
        ----------
        klass : object
            The class that will be tested.
        **kwargs
            Arguments into subclass initialization.

        """
        super().__init_subclass__(**kwargs)

        cls.klass = klass

    # /def


# /class


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Miscellaneous Utility Functions."""

__all__ = [
    # functions
    "temporary_namespace",
    "make_help_function",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from contextlib import contextmanager
from types import ModuleType

# THIRD PARTY
from astropy.utils.data import find_current_module

##############################################################################
# CODE
##############################################################################


@contextmanager
def temporary_namespace(locals_ref, keep: T.List[str] = []):
    """Temporary Namespace within ``with`` statement.

    1. copies current namespace from `locals_ref`
    2. Enters ``with`` statement
    3. restores original namespace except those specified in `keep`

    Parameters
    ----------
    module : module
        ``sys.modules[__name__]`` of module calling from.

        .. todo::

            not need to pass any module information. infer.

    keep : list, optional
        list of (str) variable names to keep.

    Yields
    ------
    locals_ref : locals
        for accessing namespace

    Warnings
    --------
    Does NOT work within functions.

    """
    original_namespace = locals_ref.copy()

    try:
        yield locals_ref
    finally:
        # difference of current and old namespace
        # without the keep keys
        drop_keys = (
            set(locals_ref.keys())
            .difference(original_namespace.keys())
            .difference(keep)
        )
        # kept values
        keep_dict = {k: locals_ref[k] for k in keep}

        # print("drop_keys", drop_keys, "keep_dict", keep_dict)

        # Restoring Namespace
        original_namespace.update(keep_dict)  # add keep values to original
        locals_ref.update(original_namespace)  # restore original (+kept)
        for key in drop_keys:  # drop unkept context-specific values
            locals_ref.pop(key)

    # /try


# -------------------------------------------------------------------


def make_help_function(
    name: str,
    module: T.Union[None, ModuleType, str] = None,
    look_for: T.Optional[str] = None,  # "Routine Listings",
    doctitle: T.Optional[str] = None,
) -> T.Callable:
    """Set docstring from module Returns section.

    Takes a helper function for a module and adds the content of the modules'
    `look_for` section. Currently only works on numpy-style docstring.

    Parameters
    ----------
    name: str
        name of function. Add "_help".
    module:
        Module
    look_for : str, optional
        The section to look for (default None)
        The section name "Routine Listings" is replaced by "Returns"
    doctitle : str, optional

    Returns
    -------
    decorator : Callable
        decorator function to change the wrapped function's docstring.

    Raises
    ------
    TypeError
        if `look_for` is not None or str

    Notes
    -----
    .. todo::

        separate the imports help function from the general helps function.
        the general help function should be similar to the find_api_page
        in astropy and the help function in the utilipy init.

    """
    if module is None:
        module = find_current_module(2)
        mod_name = module.__name__
        module_doc = module.__doc__
    elif isinstance(module, ModuleType):
        module_doc = module.__doc__
        mod_name = module.__name__
    elif isinstance(module, str):
        module_doc = module
        mod_name = find_current_module(2).__name__

    if look_for is None:
        doc = module_doc

    elif isinstance(look_for, str):
        ind = module_doc.find(look_for) + 2 * len(look_for) + 2
        end_ind = ind + module_doc[ind:].find("---")  # finding next section

        doc = module_doc[ind:end_ind]  # get section (+ next header)
        doc = "\n".join(doc.split("\n")[:-2])  # strip next header

        if look_for == "Routine Listings":  # skip 'Routine Listings' & line
            # Name = name.capitalize()
            doc = f"\nReturns\n{'-'*(len(name) + 1)}-------\n" + doc

    else:
        raise TypeError

    # TODO with FunctionType in types
    def help_function():
        print(doc)

    # help_function._doc = doc
    help_function.__name__ = f"{name}_help"
    help_function.__module__ = mod_name
    help_function.__doc__ = f"Help for {doctitle or name}.\n\n" + (doc or "")
    # /def

    return help_function


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-
# see LICENSE.rst

"""Tests for :mod:`~utilipy.utils.doc_parse_tools`."""


__all__ = [
    "test_rest_parse_tools",
    "test_numpy_parse_tools",
    "test_napolean_parse_tools",
    "test_style_store",
]


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from . import (
    test_napolean_parse_tools,
    test_numpy_parse_tools,
    test_rest_parse_tools,
    test_style_store,
)

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Data Transformer Classes."""

__author__ = "Nathaniel Starkman"
__credits__ = ["Astropy"]


__all__ = [
    "DataTransform",
    "CompositeTransform",
]


##############################################################################
# IMPORTS

# BUILT-IN
import inspect
import typing as T
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from contextlib import suppress

##############################################################################
# CODE
##############################################################################

#####################################################################
# Built-In Transform Classes


class DataTransformBase(metaclass=ABCMeta):
    """Base Class for Data Transformations.

    An object that transforms a coordinate from one system to another.
    Subclasses must implement `__call__` with the provided signature.
    They should also call this superclass's ``__init__`` in their
    ``__init__``.

    Parameters
    ----------
    fromtype : class
        The coordinate frame class to start from.
    totype : class
        The coordinate frame class to transform into.
    priority : number
        The priority if this transform when finding the shortest
        coordinate transform path - large numbers are lower priorities.
    register_graph : `TransformGraph` or `None`
        A graph to register this transformation with on creation, or
        `None` to leave it unregistered.

    """

    def __init__(self, fromtype, totype, priority=1, register_graph=None):
        if not inspect.isclass(fromtype) and fromtype is not None:
            raise TypeError("fromtype must be a class")
        if not inspect.isclass(totype) and totype is not None:
            raise TypeError("totype must be a class")

        self.fromtype = fromtype
        self.totype = totype
        self.priority = float(priority)

        if register_graph:
            # this will do the type-checking when it adds to the graph
            self.register(register_graph)
        else:
            if not (inspect.isclass(fromtype) or fromtype is None) or not (
                inspect.isclass(totype) or totype is None
            ):
                raise TypeError("fromtype and totype must be classes")

    # /def

    def register(self, graph):
        """Register this transformation into graph.

        Add this transformation to the requested Transformation graph,
        replacing anything already connecting these two data types.

        Parameters
        ----------
        graph : a TransformGraph object
            The graph to register this transformation with.

        """
        graph.add_transform(self.fromtype, self.totype, self)

    # /def

    def unregister(self, graph):
        """Remove transformation from the graph.

        Parameters
        ----------
        graph : a TransformGraph object
            The graph to unregister this transformation from.

        Raises
        ------
        ValueError
            If this is not currently in the transform graph.

        """
        graph.remove_transform(self.fromtype, self.totype, self)

    # /def

    @abstractmethod
    def __call__(self, fromdata, totype):
        """Perform the transformation from ``fromtype`` to ``totype``.

        Parameters
        ----------
        fromdata : fromtype object
            An object of class matching ``fromtype`` that is to be transformed.
        totype : object
            An object that has the attributes necessary to fully specify the
            frame.  That is, it must have attributes with names that match the
            keys of the dictionary that ``totype.get_frame_attr_names()``
            returns. Typically this is of class ``totype``, but it *might* be
            some other class as long as it has the appropriate attributes.

        Returns
        -------
        tocoord : totype object
            The new coordinate after the transform has been applied.

        """

    # /def


# /class


# -------------------------------------------------------------------


class DataTransform(DataTransformBase):
    """A transformation defined by a function.

    A transformation defined by a function that accepts a
    type and returns the transformed object.

    Parameters
    ----------
    func : callable
        The transformation function. Should have a call signature
        ``func(fromdata, *args, **kwargs)``.
    fromtype : class
        The coordinate frame class to start from.
    totype : class
        The coordinate frame class to transform into.
    priority : number
        The priority if this transform when finding the shortest
        coordinate transform path - large numbers are lower priorities.
    register_graph : `TransformGraph` or `None`
        A graph to register this transformation with on creation, or
        `None` to leave it unregistered.

    Raises
    ------
    TypeError
        If ``func`` is not callable.
    ValueError
        If ``func`` cannot accept two arguments.

    """

    def __init__(
        self,
        func: T.Callable,
        fromtype,
        totype,
        priority: int = 1,
        register_graph=None,
        func_args: T.Optional[T.Sequence] = None,
        func_kwargs: T.Optional[T.Mapping] = None,
    ):
        """Create a data transformer."""
        if not callable(func):
            raise TypeError("func must be callable")

        with suppress(TypeError):
            sig = inspect.signature(func)
            kinds = [x.kind for x in sig.parameters.values()]
            if (
                len(x for x in kinds if x == sig.POSITIONAL_ONLY) != 2
                and sig.VAR_POSITIONAL not in kinds
            ):
                raise ValueError(
                    "provided function does not accept two arguments"
                )

        self.func = func
        self.func_args = list(func_args or [])  # None -> [], keeps full
        self.func_kwargs = func_kwargs  # None -> {}, keeps full

        # TODO store these or make each time in __call__?
        self.func_sig = inspect.signature(func)
        self.func_spec = inspect.getfullargspec(func)

        super().__init__(
            fromtype, totype, priority=priority, register_graph=register_graph
        )

    # /def

    def __call__(self, fromdata, *args, _override_kws: bool = False, **kwargs):
        """Run transformation.

        Parameters
        ----------
        fromdata : Any
        *args : Any
            arguments into the transformation
        _override_kws : bool
            whether to permit the default kwargs, or completely override by any
            supplied kwargs (if any are given here). If they are permitted,
            individual arguments are still overriden by ones supplied here.
        **kwargs : Any
            keyword argument into the transformation

        Returns
        -------
        todata : Any
            The result of running `fromdata` through the transformation

        """
        # create BoundArgument
        # have None here in `fromdata` b/c will update later
        ba = self.func_sig.bind_partial(
            None, *self.func_args, **self.func_kwargs
        )
        ba.apply_defaults()  # and the defaults

        # have to override with provided arguments
        _ba = self.func_sig.bind_partial(fromdata, *args, **kwargs)
        # propagate default kwargs (unless `_override_kws`)
        vkw = self.func_spec.varkw  # the name of the varkw param
        if vkw in _ba.arguments and not _override_kws:
            keys = _ba.arguments[vkw].keys()  # do not override these keys
            _ba.arguments[vkw].update(  # update arguments
                {
                    k: ba.arguments[vkw][k]
                    for k in keys
                    if k not in keys  # filtering out locked keys
                }
            )
        ba.arguments.update(_ba.arguments)

        # call function
        todata = self.func(*ba.args, **ba.kwargs)

        totype = self.totype if self.totype is not None else type(None)

        if not isinstance(todata, totype):
            raise TypeError(
                f"the transformation function yielded {todata} but "
                f"should have been of type {totype}"
            )

        return todata

    # /def


# /class


# -------------------------------------------------------------------


class CompositeTransform(DataTransformBase):
    """A transformation from a series of single-step transformations.

    .. todo::

        - support passed arguments to steps within the transform list

    Parameters
    ----------
    transforms : sequence of `DataTransform` objects
        The sequence of transformations to apply.
    fromtype : class
        The coordinate frame class to start from.
    totype : class
        The coordinate frame class to transform into.
    priority : number
        The priority if this transform when finding the shortest
        coordinate transform path - large numbers are lower priorities.
    register_graph : `TransformGraph` or `None`
        A graph to register this transformation with on creation, or
        `None` to leave it unregistered.

    """

    def __init__(
        self,
        transforms: T.Sequence,
        fromtype,
        totype,
        priority: int = 1,
        register_graph=None,
    ):
        """Create Composite Data Transformer."""
        super().__init__(
            fromtype, totype, priority=priority, register_graph=register_graph
        )

        self.transforms = tuple(transforms)

    # /def

    def __call__(self, fromdata, *args, _override_kws: bool = False, **kwargs):
        """Run transformation.

        Parameters
        ----------
        fromdata : Any
        *args : Any
            arguments into the first transformation
        _override_kws : bool
            whether to permit the default kwargs in the first transformation,
            or completely override by any supplied kwargs (if any are given
            here). If they are permitted, individual arguments are still
            overriden by ones supplied here.
        **kwargs : Any
            keyword argument into the first transformation

        Returns
        -------
        todata : Any
            The result of running `fromdata` through the transformation series
            listed in ``self.transforms``

        """
        todata = fromdata
        for i, t in enumerate(self.transforms):
            if i == 0:  # TODO what if doesn't accept args/kwargs?
                todata = t(
                    todata, *args, _override_kws=_override_kws, **kwargs
                )
            else:
                todata = t(todata)

        return todata

    # /def


# /class


#####################################################################
# Default Transformation Set
# TODO seed many more basics
#    - can do astropy and numpy ones
# AND do in separate file?


_default_xfm_set = defaultdict(
    dict,
    {
        None: {  # TODO replace this with an Any-to-None function
            k: DataTransform(lambda x: None, k, None)
            for k in (str, list, tuple, dict)
        },
        tuple: {  # TODO replace this with a many-to-tuple function
            k: DataTransform(lambda x: tuple(x), k, tuple)
            for k in (str, list, dict)
        },
        list: {  # TODO replace this with a many-to-list function
            k: DataTransform(lambda x: list(x), k, list)
            for k in (str, tuple, dict)
        },
        str: {  # TODO replace this with a many-to-str function
            k: DataTransform(lambda x: str(x), k, str)
            for k in (tuple, list, dict)
        },
        dict: {
            **{  # TODO replace this with a many-to-dict function
                k: DataTransform(lambda x: dict(x), k, dict)
                for k in (tuple, list, dict)
            },
            **{
                None: DataTransform(lambda x: dict(), None, dict)
            },  # None to dict
        },
    },
)


# # # map class names to colorblind-safe colors
# trans_to_color = OrderedDict()
# # trans_to_color[AffineTransform] = '#555555'  # gray
# trans_to_color[DataTransform] = "#783001"  # dark red-ish/brown
# # trans_to_color[FunctionTransformWithFiniteDifference] = '#d95f02'  # red-ish
# # trans_to_color[StaticMatrixTransform] = '#7570b3'  # blue-ish
# # trans_to_color[DynamicMatrixTransform] = '#1b9e77'  # green-ish


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : plotly imports
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

"""Base set of imports for :mod:`~plotly`.

Only imports if plotly is installed

Routine Listings
----------------
plotly : imports

    - plotly
    - express -> px
    - graph_objs -> go
    - io -> pio
    - subplots -> make_subplots

plotly_imports_help


References
----------
Plotly reference [#]_.

.. [#] plotly: Collaborative data science, Plotly Technologies (2015),
    (https://plot.ly)

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "plotly_imports_help",
    "HAS_PLOTLY",
]


##############################################################################
# HELPER FUNCTIONS

# PROJECT-SPECIFIC
from utilipy.imports import conf
from utilipy.utils import make_help_function

##############################################################################
# IMPORTS

try:

    import plotly

except ImportError:
    import warnings

    warnings.warn("Cannot import plotly")

    HAS_PLOTLY: bool = False

else:

    HAS_PLOTLY: bool = True

    import plotly.express as px
    import plotly.graph_objects as go
    import plotly.io as pio
    from plotly.subplots import make_subplots

    __all__ += ["plotly", "px", "go", "pio", "make_subplots"]

# /if


##############################################################################
# Printing Information

plotly_imports_help = make_help_function(
    "plotly", __doc__, look_for="Routine Listings"
)


if conf.verbose_imports:
    plotly_imports_help()


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Imports Sets.

Most python scripts have a large number of lines dedicated to just
importing various modules and functions. While it is good to be explicit,
sometimes it's nice to just start coding. Also, its a pain to copy the
same set of imports between scripts, and scrolling through all the imports
to reach the start of the code is aggravating.

This module provides a variety of files that can be ``*``-imported to provide
basic set of imports so that you can just get started coding.


The provided quick imports are ``base_imports``, ``extended_imports``,
``astropy_imports``, ``matplotlib_imports``, ``galpy_imports`` and
``amuse_imports``.

The files will print an import summary. To prevent this summary, set
``verbose_imports=False`` in the configuration file in your
home directory. Each of the imports also provides a helper
function that will print out the import summary.


Package-Specific Import Sets
----------------------------

Most packages have a standard set of imports. Often these imports are not
populated directly in the top-level namespace of the package, so that a
`from package import *` would suffice to import the standard set of imports.
The purpose of this module is to provide such sets of standard imports,
so that, for the `matplotlib` package as an example, the following are
imported with one line

    `from utilipy.imports.matplotlib_imports import *`

    - pyplot->plt
    - matplotlib->mpl, .cm, .colors
    - mpl_toolkits.mplot3d.Axes3D



"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "conf",
    "use_import_verbosity",
]


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from .setup_package import conf, use_import_verbosity

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Catalog x-matching on data fields.

Built from the extensive x-matching program in Jo Bovy's ``gaia_tools``,
but reconfigured to use the Astropy framework as much as possible.

"""

__author__ = "Nathaniel Starkman"
__credits__ = ["Jo Bovy", "Henry Leung"]


__all__ = [
    "indices_xmatch_fields",
    "xmatch_fields",
    "xmatch",
    "non_xmatched",
]


##############################################################################
# IMPORTS

# BUILT-IN
import functools
import itertools
import typing as T

# THIRD PARTY
import numpy as np

# PROJECT-SPECIFIC
from .decorators import idxDecorator

# from .xfm import data_graph as old_data_graph, DataTransform


##############################################################################
# PARAMETERS

_TBL_TYPE = np.recarray
_FIELDS_TYPE = T.List[str]
_IDX_TYPE = np.array
_IDXS_TYPE = T.List[_IDX_TYPE]
_INFO_TYPE = T.Dict[str, T.Sequence]


##############################################################################
# CODE
##############################################################################


@idxDecorator(as_ind=False)
def _indices_equality_match_on_catalog(
    catalog: _TBL_TYPE, other: _TBL_TYPE, fields: _FIELDS_TYPE
) -> _IDX_TYPE:
    """Indices of catalog data field(s) to match against a source catalog.

    This function is for discrete-valued data, such as tags.
    Note that this only matches `other` against `catalog`, meaning
    `catalog` can still have values which do not appear in `other`
    For coordinates, see `~indices_xmatch_coords`.

    This match is done on all the fields simultaneously, so in terms or
    a 2D array, two rows are considered a match only if all the values
    in the columns `fields` match.

    Parameters
    ----------
    catalog : recarray
        the source catalog against which the `other` catalog is matched.
    other : recarray
        match this against `catalog`
    fields : list
        List of fields on which to match.
        ex, ["color", "location"] where both `catalog` and `other` have
        those columns, hopefully with some matching values.

    Returns
    -------
    idx : ndarray
        indices into `other` such that only has values in `fields` that
        are in `catalog`.

    Notes
    -----
    .. todo::

        try more axes tricks to avoid loops.

    """
    # uniques for each field
    uns: T.Tuple[np.array]
    uns = (np.unique(np.array(catalog[n])) for n in fields)

    # indices into cat1
    # loop over fields  TODO vectorize ?
    idxs = (other[n] == un[:, None] for n, un in zip(fields, uns))
    # get combined index
    idx: _IDX_TYPE
    idx = np.sum(functools.reduce(np.logical_and, idxs), axis=0, dtype=bool)

    return idx


# /def


def indices_xmatch_fields(
    catalog: _TBL_TYPE, *others: _TBL_TYPE, fields: _FIELDS_TYPE
) -> T.Tuple[_IDXS_TYPE, _INFO_TYPE]:
    """Indices of xmatch of catalogs' data field(s) against a source catalog.

    This function is for discrete-valued data, such as tags.
    For coordinates, see `~indices_xmatch_coords`.

    This match is done on all the fields simultaneously, so in terms or
    a 2D array, two rows are considered a match only if all the values
    in the columns `fields` match.

    Parameters
    ----------
    catalog : Table or recarray
        The source catalog against which the `others` catalogs are matched.
    *others : Table or recarray
        match these against `catalog`
    fields : list
        List of fields on which to match.
        ex, ["color", "location"] where both `catalog` and `other` have
        those columns, hopefully with some matching values.

    Returns
    -------
    idxs : list of ndarrays
        each element of list is the x-match indices into the ith catalog,
        starting with `catalog`. So the i=0 index list is for `catalog` and
        the i=1 index list is the x-match indices for the first table
        in `others`.
    info : dict
        Useful information. Nothing yet.

    Notes
    -----
    .. todo::

        try more axes tricks to avoid loops.

    """
    idxs: _IDXS_TYPE = [
        _indices_equality_match_on_catalog(catalog, c, fields=fields)
        for c in others
    ]  # TODO with numpy
    # and need to do once in reverse on the source catalog
    catalog_idx: _IDX_TYPE = _indices_equality_match_on_catalog(
        others[1][idxs[0]], catalog, fields
    )
    idxs.insert(0, catalog_idx)

    info: _INFO_TYPE = {}

    return idxs, info


# /def


def xmatch_fields(
    catalog: _TBL_TYPE, *others: _TBL_TYPE, fields: _FIELDS_TYPE
) -> T.Tuple[T.List[T.Any], _INFO_TYPE]:
    """Cross-match catalogs' data field(s) against a source catalog.

    This function is for discrete-valued data, such as tags.
    For coordinates, see `~indices_xmatch_coords`.

    This match is done on all the fields simultaneously, so in terms or
    a 2D array, two rows are considered a match only if all the values
    in the columns `fields` match.

    Parameters
    ----------
    catalog : Table or recarray
        The source catalog against which the `others` catalogs are matched.
        fields for which there are no matches are also filtered.
    *others : Table or recarray
        match these against `catalog`
    fields : list
        List of fields on which to match.
        ex, ["color", "location"] where both `catalog` and `other` have
        those columns, hopefully with some matching values.

    Returns
    -------
    idxs : list of ndarrays
        each element of list is the x-match indices into the ith catalog,
        starting with `catalog`. So the i=0 index list is for `catalog` and
        the i=1 index list is the x-match indices for the first table
        in `others`,


    Notes
    -----
    .. todo::

        try more axes tricks to avoid loops.

    """
    if not isinstance(fields, list):  # TODO less draconian
        raise TypeError("must be a list")
    if len(others) == 0:
        raise ValueError(
            "Must have at least one catalog against-which to xmatch."
        )

    # iterate through all catalogs, making sure has the field name
    # iterators are the catalog number and field names
    # these are itered together to generate all list combinations.
    for i, field in itertools.product(range(len(others) + 1), fields):
        try:
            if i == 0:
                catalog[field]
            else:
                others[i][field]
        except Exception as e:  # TODO better error message
            print(f"need to have {field} in catalog {i}")
            raise e

    # ------------------------------

    idxs: _IDXS_TYPE
    info: _INFO_TYPE
    idxs, info = indices_xmatch_fields(catalog, *others, fields=fields)

    cat_matches = [catalog[idxs[0]]] + [
        c[idx] for c, idx in zip(others, idxs[1:])
    ]

    info.update({"idxs": idxs})

    return cat_matches, info


xmatch_tags = xmatch_fields  # just another name

# /def


##############################################################################
# General Cross-Match

# @data_graph.decorate(cat1=np.recarray, cat2=np.recarray)
def xmatch(
    *catalogs,
    match_fields: T.Sequence,
) -> T.Tuple[T.List[T.Any], _INFO_TYPE]:
    """Cross-match two catalogs.

    .. todo::

        - find duplicates before matching ?
        - allow more catalogs to be xmatched

    Parameters
    ----------
    cat1, cat2: Any
        the two catalogs to crossmatch
        must be convertible to recarray
    match_fields : str, optional
        data tag on which to additionally cross-match, default None.
        this also works for any discrete-valued data column.

    References
    ----------
    https://github.com/jobovy/gaia_tools/

    """
    cat_matches, info = xmatch_fields(*catalogs, fields=match_fields)

    return cat_matches, info


# /def

# -------------------------------------------------------------------


def non_xmatched(
    catalog1: T.Sequence,
    catalog2: T.Sequence,
    indices1: T.Sequence,
    indices2: T.Sequence,
):
    """Find non cross-matched catalog components.

    Parameters
    ----------
    catalog1, catalog2 : Sequence
        the catalogs
    indices1, indices2 : Sequence
        indices into catalogs for the x-match.
        :func:`~starkman_thesis.utils.data.xmatch.xmatch_indices_coords` output
        or in :func:`~starkman_thesis.utils.data.xmatch.xmatch_coords` info

    Returns
    -------
    catalog1_matches, catalog2_matches : catalog input types
        the x-matched catalogs
    info : dict
        Useful information.

            - nindices1 : indices into `catalog1` not x-matched.
            - nindices1 : indices into `catalog2` not x-matched.

    """
    # index arrays of length catalog, for getting non-matches
    c1idx = np.arange(len(catalog1))  # catalog1 array
    c2idx = np.arange(len(catalog2))  # catalog2 array
    # bool array of non-matches
    nindices1 = np.where(~np.in1d(c1idx, indices1))[0]
    nindices2 = np.where(~np.in1d(c2idx, indices2))[0]

    ninfo = {
        "nindices1": nindices1,
        "nindices2": nindices2,
    }  # non-match info dict

    return (catalog1[nindices1], catalog2[nindices2]), ninfo


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Fitting utilities."""

__all__ = [
    # modules
    "lmfit_utils",
    # functions
    "scipy_residual_to_lmfit",
]


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from . import lmfit_utils
from .lmfit_utils import scipy_residual_to_lmfit

# from .astropy_decorator import scipy_function_to_astropy_model


##############################################################################
# END



# -*- coding: utf-8 -*-


"""Decorators."""


__all__ = [
    # modules
    "baseclass",
    "docstring",
    "func_io",
    "code_dev",
    # defined here
    "store_function_input",
    "add_folder_backslash",
    # baseclass
    "DecoratorBaseClass",
    "classy_decorator",
    # data-type decorators
    "dtypeDecorator",
    "dtypeDecoratorMaker",
    "intDecorator",
    "floatDecorator",
    "strDecorator",
    "boolDecorator",
    "ndarrayDecorator",
    "ndfloat64Decorator",
    # convenience
    "functools",
    "inspect",
    "wraps",
]


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from . import baseclass, code_dev, docstring, func_io
from .baseclass import DecoratorBaseClass, classy_decorator
from .func_io import (
    boolDecorator,
    dtypeDecorator,
    dtypeDecoratorMaker,
    floatDecorator,
    intDecorator,
    ndarrayDecorator,
    ndfloat64Decorator,
    store_function_input,
    strDecorator,
)
from utilipy.utils import functools, inspect
from utilipy.utils.functools import wraps

from .func_io import add_folder_backslash  # dtype; standard types; numpy

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Utilities for :mod:`~utilipy.data_utils`."""


__all__ = [
    "intermix_arrays",
    "make_shuffler",
    "get_path_to_file",
]


##############################################################################
# IMPORTS

# BUILT-IN
import pathlib
import typing as T

# THIRD PARTY
import numpy as np
from astropy.utils.data import get_pkg_data_filenames

##############################################################################
# PARAMETERS


##############################################################################
# CODE
##############################################################################


# TODO stride option for block sizes
def intermix_arrays(*arrs: T.Sequence, axis=-1):
    """Intermix arrays.

    Parameters
    ----------
    *arrs : Sequence
    axis : int, optional

    Return
    ------
    arr : Sequence

    Examples
    --------
    Mix single scalar array (does nothing)

        >>> x = np.arange(5)
        >>> intermix_arrays(x)
        array([0, 1, 2, 3, 4])

    Mix two scalar arrays

        >>> y = np.arange(5, 10)
        >>> intermix_arrays(x, y)
        array([0, 5, 1, 6, 2, 7, 3, 8, 4, 9])

    Mix multiple scalar arrays

        >>> z = np.arange(10, 15)
        >>> intermix_arrays(x, y, z)
        array([ 0,  5, 10,  1,  6, 11,  2,  7, 12,  3,  8, 13,  4,  9, 14])

    Mix single ND array

        >>> xx = np.c_[x, y]
        >>> intermix_arrays(xx)
        array([[0, 1, 2, 3, 4],
               [5, 6, 7, 8, 9]])

    Mix two ND arrays

        >>> yy = np.c_[z, np.arange(15, 20)]
        >>> intermix_arrays(xx, yy)
        array([[ 0, 10,  1, 11,  2, 12,  3, 13,  4, 14],
               [ 5, 15,  6, 16,  7, 17,  8, 18,  9, 19]])

    """
    shape = list(np.asanyarray(arrs[0]).shape[::-1])
    shape[axis] *= len(arrs)

    return np.array(arrs).T.flatten().reshape(shape)


# /def


# -------------------------------------------------------------------


def make_shuffler(
    length: int, rng=None
) -> T.Tuple[T.Sequence[int], T.Sequence[int]]:
    """
    Shuffle and Unshuffle arrays.

    Parameters
    ----------
    length : int
        Array length for which to construct (un)shuffle arrays.
    rng : :class:`~numpy.random.Generator` instance, optional
        random number generator.

    Returns
    -------
    shuffler : `~numpy.ndarray`
        index array that shuffles any array of size `length` along
        a specified axis
    undo : `~numpy.ndarray`
        index array that undoes above, if applied identically.

    """
    if rng is None:
        try:
            rng = np.random.default_rng()
        except AttributeError:
            rng = np.random

    # start with index array
    shuffler = np.arange(length)
    # now shuffle array (in-place)
    rng.shuffle(shuffler)

    # and construct the unshuffler
    undo = shuffler.argsort()

    return shuffler, undo


# /def


# -------------------------------------------------------------------


def get_path_to_file(*data_name: str, package=None):
    """Get path to file.

    Similar to :func:`~astropy.utils.data.get_pkg_data_filename`, but only
    gets the path to the file, does not check if the file exists.
    The package directory must exist.

    Parameters
    ----------
    *data_name : tuple of str
        the path, without folder delimiters (which are platform specific)
        relative to `package`.

        One of the following:

        * The name of a directory included in the source
          distribution.  The path is relative to the module
          calling this function.  For example, if calling from
          ``astropy.pkname``, use ``'data'`` to get the
          files in ``astropy/pkgname/data``.
        * Remote URLs are not currently supported.

    package : str, optional
        If specified, look for a file relative to the given package, rather
        than the default of looking relative to the calling module's package.

    Returns
    -------
    filename : str
        A file path on the local file system corresponding to the data
        requested in ``data_name``.

    """
    # TODO more robust system
    fps = list(get_pkg_data_filenames(".", package=package))
    filename = pathlib.Path(fps[0]).parent.joinpath(*data_name)

    return str(filename)


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests :mod:`~utilipy.utils.functools`."""


__all__ = [
    "test_WRAPPER_ASSIGNMENTS",
    "test_SIGNATURE_ASSIGNMENTS",
    "test_WRAPPER_UPDATES",
    # "test_update_wrapper_signature_true",
    # "test_update_wrapper_signature_false",
    # "test_update_wrapper_signature_None",
    # "test_update_wrapper_signature_Signature",
    # "test_update_wrapper_signature_FullerSignature",
    # "test_update_wrapper_signature_notSignature",
]


###############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from utilipy.utils import functools, inspect

###############################################################################
# PARAMETERS


def test_WRAPPER_ASSIGNMENTS():
    """Test `~utilipy.utils.functools.WRAPPER_ASSIGNMENTS`."""
    assert functools.WRAPPER_ASSIGNMENTS == (
        "__module__",
        "__name__",
        "__qualname__",
        "__doc__",
        "__annotations__",
    )


# /def


def test_SIGNATURE_ASSIGNMENTS():
    """Test `~utilipy.utils.functools.SIGNATURE_ASSIGNMENTS`."""
    assert functools.SIGNATURE_ASSIGNMENTS == (
        "__kwdefaults__",
        "__annotations__",
    )


# /def


def test_WRAPPER_UPDATES():
    """Test `~utilipy.utils.functools.WRAPPER_UPDATES`."""
    assert functools.WRAPPER_UPDATES == ("__dict__",)


# /def

##############################################################################
# CODE
##############################################################################

#############################################################################
# make_function


def test_make_function():
    """Test :func:`~utilipy.utils.functools.make_function`."""
    # test function, to copy
    def _test_func(x=None):
        return x, 2

    # /def

    sig = inspect.signature(_test_func)
    params = list(sig.parameters.values())
    params[0] = params[0].replace(default=2)
    sig = sig.replace(parameters=params)

    test_func = functools.make_function(
        _test_func.__code__,
        globals_=globals(),
        name="made_func",
        signature=sig,
    )

    assert test_func.__name__ == "made_func"

    assert _test_func() == (None, 2)
    assert test_func() == (2, 2)

    assert test_func(1) == (1, 2)


# /def

# --------------------------------------------------------------------------


##############################################################################
# copy_function


def test_copy_function():
    """Test `~utilipy.utils.functools.copy_function`."""
    # test function
    def test_func(x: float, y, a=2, b=3, *args, p="L", q="M", **kwargs):
        return x, y, a, b, args, p, q, kwargs

    tfc = functools.copy_function(test_func)

    # ------------------------------------
    # test properties

    sep_tests = (  # require separate tests
        "__call__",
        "__delattr__",
        "__dir__",
        "__eq__",
        "__format__",
        "__ge__",
        "__get__",
        "__getattribute__",
        "__gt__",
        "__hash__",
        "__init__",
        "__le__",
        "__lt__",
        "__ne__",
        "__reduce__",
        "__reduce_ex__",
        "__repr__",
        "__setattr__",
        "__sizeof__",
        "__str__",
    )

    for attr in set(dir(test_func)).difference(sep_tests):
        assert getattr(tfc, attr) == getattr(test_func, attr), attr

    # __call__
    ba = inspect.signature(test_func).bind(
        0,
        [1, 1.5],
        2.1,
        3.1,
        4.0,
        4.33,
        4.66,
        p=5,
        q=[6, 6.5],
        r=7,
        s=[8, 8.5],
    )
    assert tfc.__call__(*ba.args, **ba.kwargs) == test_func.__call__(
        *ba.args, **ba.kwargs
    )

    # __delattr__
    assert hasattr(tfc, "attr_to_del") == hasattr(test_func, "attr_to_del")
    tfc.attr_to_del = "delete me"
    test_func.attr_to_del = "delete me"
    assert hasattr(tfc, "attr_to_del") == hasattr(test_func, "attr_to_del")
    assert tfc.attr_to_del == test_func.attr_to_del  # test equality
    tfc.__delattr__("attr_to_del")
    test_func.__delattr__("attr_to_del")
    assert hasattr(tfc, "attr_to_del") == hasattr(test_func, "attr_to_del")

    # __setattr__
    assert hasattr(tfc, "attr_to_del") == hasattr(test_func, "attr_to_del")
    tfc.__setattr__("attr_to_del", "delete me")
    test_func.__setattr__("attr_to_del", "delete me")
    assert hasattr(tfc, "attr_to_del") == hasattr(test_func, "attr_to_del")
    assert tfc.attr_to_del == test_func.attr_to_del  # and be equal
    del tfc.attr_to_del, test_func.attr_to_del
    assert hasattr(tfc, "attr_to_del") == hasattr(test_func, "attr_to_del")

    # __dir__
    assert set(tfc.__dir__()) == set(test_func.__dir__())

    # __getattribute__
    assert set(tfc.__getattribute__("__dir__")()) == set(
        test_func.__getattribute__("__dir__")()
    )

    # __eq__, __ge__, __gt__, '__le__', '__lt__', '__ne__'
    for attr in ("__eq__", "__ge__", "__gt__", "__le__", "__lt__", "__ne__"):
        assert tfc.__getattribute__(attr)(tfc) == test_func.__getattribute__(
            attr
        )(test_func)
        assert tfc.__getattribute__(attr)(
            test_func
        ) == test_func.__getattribute__(attr)(tfc)

    # __format__
    # TODO

    # __get__
    # TODO

    # __hash__
    # TODO

    # __init__
    # TODO

    # __reduce__, __reduce_ex__ TypeError: can't pickle function objects

    # __repr__
    # TODO, right now they are the same. Problem?

    # __sizeof__
    assert tfc.__sizeof__() == test_func.__sizeof__()

    # __str__
    assert tfc.__str__() != test_func.__str__()
    assert tfc.__str__().split(" ")[:-1] == test_func.__str__().split(" ")[:-1]

    # ------------------------------------
    # test calling

    # TODO

    return


# /def


# ##############################################################################
# # update_wrapper


# def test_update_wrapper_defaults():

#     # ----------------------------------------------
#     # basic decorator

#     def decorator(function):
#         def wrapper(*args, **kwargs):
#             return function(*args, **kwargs)

#         return wrapper

#     # /def

#     for name in _test_functools_util.__all__:
#         func = getattr(_test_functools_util, name)
#         f_sig = inspect.signature(func)

#         dec_func = functools.update_wrapper(decorator(func), func)
#         df_sig = inspect.signature(dec_func)

#         # assigned
#         uwsig = inspect.signature(functools.update_wrapper)
#         assigned = uwsig.parameters["assigned"].default
#         for attr in assigned:
#             assert getattr(dec_func, attr) == getattr(func, attr)

#         # # updated  # doesn't work b/c stuff added to __dict__ after
#         # updated = uwsig.parameters["updated"].default
#         # for attr in updated:
#         #     assert getattr(dec_func, attr) == getattr(func, attr)

#         assert not hasattr(func, "__signature__")
#         assert hasattr(dec_func, "__signature__")
#         assert dec_func.__signature__ == df_sig

#         # signature
#         for attr in functools.SIGNATURE_ASSIGNMENTS:
#             assert getattr(dec_func, attr) == getattr(func, attr)

#         assert df_sig == f_sig

#         # wrapped
#         assert hasattr(dec_func, "__wrapped__")
#         assert dec_func.__wrapped__ == func

#         # TODO more tests

#     # /for

#     # ----------------------------------------------
#     # decorator with kwargs

#     def decorator(function):
#         def wrapper(*args, kw1="test", **kwargs):
#             return function(*args, **kwargs)

#         return wrapper

#     # /def


#     for name in _test_functools_util.__all__:
#         func = getattr(_test_functools_util, name)

#         f_sig = inspect.signature(func)

#         dec_func = functools.update_wrapper(decorator(func), func)
#         df_sig = inspect.signature(dec_func)

#         # assigned
#         uwsig = inspect.signature(functools.update_wrapper)
#         assigned = uwsig.parameters["assigned"].default
#         for attr in assigned:
#             assert getattr(dec_func, attr) == getattr(func, attr)

#         # # updated  # doesn't work b/c stuff added to __dict__ after
#         # updated = uwsig.parameters["updated"].default
#         # for attr in updated:
#         #     assert getattr(dec_func, attr) == getattr(func, attr)

#         assert not hasattr(func, "__signature__")
#         assert hasattr(dec_func, "__signature__")
#         assert dec_func.__signature__ == df_sig

#         # signature
#         for attr in functools.SIGNATURE_ASSIGNMENTS:
#             dfa = {k: v for k, v in getattr(dec_func, attr).items() if k != "kw1"}
#             assert dfa == (getattr(func, attr) or {})

#         # same signature, except for 'kw1'
#         indx = list(df_sig.parameters.keys()).index("kw1")
#         sig = list(df_sig.parameters.values())
#         del sig[indx]
#         assert sig == list(f_sig.parameters.values())

#         # signature adds 'kw1' after *args
#         indx = list(df_sig.parameters.keys()).index("kw1")
#         list(df_sig.parameters.values())[indx - 1].kind == inspect._VAR_POSITIONAL

#         # wrapped
#         assert hasattr(dec_func, "__wrapped__")
#         assert dec_func.__wrapped__ == func

#         # TODO more tests

#     # /for

# # ----------------------------------------------
# # decorator with args


# def decorator(function):
#     def wrapper(x, *args, **kwargs):
#         return function(*args, **kwargs)

#     return wrapper


# # /def

# # TODO

# # ----------------------------------------------

# return


# # /def


# # -----------------------------------------------------------------------------


# @pytest.mark.skip(reason="TODO")
# def test_update_wrapper_signature_true():
#     """Test the case when `signature`=True."""
#     return


# # /def


# @pytest.mark.skip(reason="TODO")
# def test_update_wrapper_signature_false():
#     """Test the case when `signature`=False."""
#     return


# # /def


# @pytest.mark.skip(reason="TODO")
# def test_update_wrapper_signature_None():
#     """Test the case when `signature`=None."""
#     return


# # /def


# @pytest.mark.skip(reason="TODO")
# def test_update_wrapper_signature_Signature():
#     """Test the case when `signature`is an inspect.Signature object."""
#     return


# # /def


# @pytest.mark.skip(reason="TODO")
# def test_update_wrapper_signature_FullerSignature():
#     """Test the case when `signature`is an FullerSignature object."""
#     return


# # /def


# @pytest.mark.skip(reason="TODO")
# def test_update_wrapper_signature_notSignature():
#     """Test the case when `signature`is not a Signature object.
#     This should raise an error.

#     """
#     return


# # /def


# ###############################################################################
# # wraps


# ###############################################################################
# # END



# -*- coding: utf-8 -*-

"""Test data_util :mod:`~utilipy.data_utils.decorators`."""


__all__ = [
    "test_make_shuffler",
    "test_intermix_arrays",
]


##############################################################################
# IMPORTS

# THIRD PARTY
import numpy as np

# PROJECT-SPECIFIC
from utilipy.data_utils import utils

##############################################################################
# PARAMETERS

x = np.arange(2)
y = np.arange(2) + 2

z = np.c_[x, y].T


##############################################################################
# TESTS
##############################################################################


def test_make_shuffler():
    """Test :func:`~utilipy.data_utils.utils.shuffle`."""
    arr = np.arange(100)

    shuffler, undo = utils.make_shuffler(len(arr))

    # shuffling an identity index array yields the shuffler
    assert all((arr[shuffler] - shuffler) == 0.0)

    # unshuffling the shuffler is the identity index array
    assert all(shuffler[undo] == arr)

    # unshufffling works
    assert all(arr[shuffler][undo] == arr)

    return


# /def


# ------------------------------------------------------------------------


def test_intermix_arrays():
    """Test :func:`~utilipy.data_utils.utils.intermix_arrays`."""
    x = np.arange(5)
    y = np.arange(5, 10)
    z = np.arange(10, 15)
    xx = np.c_[x, y]
    yy = np.c_[z, np.arange(15, 20)]

    # Mix single scalar array
    m = utils.intermix_arrays(x)
    expected = np.array([0, 1, 2, 3, 4])

    assert np.all(m == expected)

    # Mix two scalar arrays
    m = utils.intermix_arrays(x, y)
    expected = np.array([0, 5, 1, 6, 2, 7, 3, 8, 4, 9])

    assert np.all(m == expected)

    # Mix multiple scalar arrays
    m = utils.intermix_arrays(x, y, z)
    expected = np.array([0, 5, 10, 1, 6, 11, 2, 7, 12, 3, 8, 13, 4, 9, 14])

    assert np.all(m == expected)

    # Mix single ND array
    m = utils.intermix_arrays(xx)
    expected = np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]])

    assert np.all(m == expected)

    # Mix two ND arrays
    m = utils.intermix_arrays(xx, yy)
    expected = np.array(
        [
            [0, 10, 1, 11, 2, 12, 3, 13, 4, 14],
            [5, 15, 6, 16, 7, 17, 8, 18, 9, 19],
        ]
    )

    assert np.all(m == expected)

    # Axis arguments
    m = utils.intermix_arrays(xx, yy, axis=0)
    expected = np.array(
        [
            [0, 10, 1, 11, 2],
            [12, 3, 13, 4, 14],
            [5, 15, 6, 16, 7],
            [17, 8, 18, 9, 19],
        ]
    )

    assert np.all(m == expected)


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import

# def get_package_data():
#     """Get Package Data for utilipy.scripts"""
#     return {'utilipy.scripts': ['data/*']}
# # /def



# -*- coding: utf-8 -*-

"""Utilities."""


__all__ = [
    # modules
    "collections",
    "logging",
    "exceptions",
    "functools",
    "inspect",
    "metaclasses",
    "misc",
    "pickle",
    "string",
    "typing",
    # classes and functions
    "LogPrint",
    "LogFile",
    "ObjDict",
    "WithDocstring",
    "WithMeta",
    "WithReference",
    "temporary_namespace",
    "make_help_function",
]

__all_top_imports__ = (  # TODO deprecate
    "collections",
    # "doc_parse_tools",
    "logging",
    "exceptions",
    "functools",
    "inspect",
    "metaclasses",
    "pickle",
    "string",
    "typing",
)


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
# import modules
from . import inspect  # doc_parse_tools,
from . import (
    collections,
    exceptions,
    functools,
    logging,
    metaclasses,
    misc,
    pickle,
    string,
    typing,
)
from .collections import ObjDict, WithDocstring, WithMeta, WithReference
from .logging import LogFile, LogPrint
from .misc import make_help_function, temporary_namespace

##############################################################################
# CODE
##############################################################################


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Test functions for :mod:`~matplotlib` imports."""


##############################################################################


def test_import_matplotlib():
    """Test import matplotlib."""
    from utilipy.imports import matplotlib_imports as imports

    for obj in ("mpl", "plt", "cm", "colors"):
        assert hasattr(imports, obj)


# /def


# -------------------------------------------------------------------


def test_import_matplotlib_has_ipython():
    """Test matplotlib imports when in IPython environment.

    combines with test_import_matplotlib to test all imports.

    """
    from utilipy.imports import matplotlib_imports as imports

    try:
        get_ipython()
        if get_ipython() is None:  # double checking
            raise NameError

    except NameError:
        pass

    else:

        imports.configure_matplotlib

    return


# /def

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Data Utilities."""


__all__ = [
    # modules
    "crossmatch",
    "decorators",
    "select",
    "fitting",
    "utils",
    "xfm",
    # functions
    "idxDecorator",
    # data transformation graph
    "data_graph",
    "TransformGraph",
    "DataTransform",
    # xmatch
    "indices_xmatch_fields",
    "xmatch_fields",
    "xmatch",
    "non_xmatched",
    # utils
    "make_shuffler",
    "get_path_to_file",
]


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from . import crossmatch, decorators, fitting, select, utils, xfm
from .crossmatch import (
    indices_xmatch_fields,
    non_xmatched,
    xmatch,
    xmatch_fields,
)
from .decorators import idxDecorator
from .select import *  # noqa
from .utils import get_path_to_file, make_shuffler
from .xfm import DataTransform, TransformGraph, data_graph

# -------------------------------------------------------------------
# __ALL__

__all__ += select.__all__


##############################################################################
# DONE



# -*- coding: utf-8 -*-
# see LICENSE.rst

"""Test for :mod:`~utilipy.imports`."""


__all__ = [
    "test_amuse_imports",
    "test_astropy_imports",
    "test_base_imports",
    "test_extended_imports",
    "test_galpy_imports",
    "test_matplotlib_imports",
]


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from . import (
    test_amuse_imports,
    test_astropy_imports,
    test_base_imports,
    test_extended_imports,
    test_galpy_imports,
)

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for :mod:`~utilipy.utils.collections`."""


__all__ = [
    "test_WithDocstring",
    "test_WithMeta",
    "test_WithReference",
    # -------
    "test_MetaDataBase",
    "test_ReferenceBase",
    # -------
    "test_ObjDict_creation",
    "test_ObjDict_getitem",
    "test_ObjDict__setitem__",
    "test_ObjDict__getattr__",
    "test_ObjDict__setattr__",
    "test_ObjDict__repr__",
    "test_ObjDict_values",
    "test_ObjDict_items",
    "test_ObjDict_fromkeys",
    "test_ObjDict_keyslist",
    "test_ObjDict__reduce__",
    "test_ObjDict__setstate__",
    "test_ObjDict_dump",
    "test_ObjDict_save",
    "test_ObjDict_load",
    "test_ObjDict_print",
]


###############################################################################
# IMPORTS

# BUILT-IN
import os.path
import tempfile
from collections import OrderedDict

# THIRD PARTY
import pytest

# PROJECT-SPECIFIC
from utilipy.utils import collections
from utilipy.utils.collections import ObjDict

###############################################################################
# PARAMETERS


###############################################################################
# CODE
###############################################################################

###############################################################################
# WithDdocstring


def test_WithDocstring():
    """Test :class:`~utilipy.utils.collections.WithDdocstring`."""
    doc = "this docstring has been added"

    # ------------------------------------------------------------------------
    # test integer

    x = collections.WithDocstring(2, doc=doc)

    assert x.__doc__ == doc  # test has docstring
    assert x.__wrapped__.__doc__ != doc  # test wrapped object does not

    # test operations do not carry through.
    y = x + 2
    assert y == 4
    assert y.__doc__ != doc

    # --------------------------------------------------
    # test something with docstring

    class Test(object):
        """Docstring for Test."""

        def __init__(self):
            super().__init__()

    # start with object
    test = Test()
    assert test.__doc__ == """Docstring for Test."""  # docstring matches

    # wrap & test
    wrapped = collections.WithDocstring(test, doc=doc)

    assert wrapped.__doc__ == doc  # test has docstring
    assert wrapped.__wrapped__.__doc__ == """Docstring for Test."""

    return


# /def


###############################################################################
# WithMeta


def test_WithMeta():
    """Test :class:`~utilipy.utils.collections.WithMeta`."""
    wm = collections.WithMeta(2, a="A", b="B")

    assert wm == 2

    # test meta
    assert hasattr(wm, "meta")
    assert isinstance(wm.meta, OrderedDict)
    assert wm.meta["a"] == "A"
    assert wm.meta["b"] == "B"

    with pytest.raises(KeyError):  # it shouldn't have this
        assert wm.meta["c"]

    # --------------------------------------------------
    # TODO test something with existing meta

    return


# /def


###############################################################################
# WithReference


def test_WithReference():
    """Test :class:`~utilipy.utils.collections.WithReference`."""
    doc = "this docstring has been added"
    ref = "Reference"

    wr = collections.WithReference(2, doc=doc, reference=ref, a="A", b="B")

    assert wr == 2

    # test docstring
    assert wr.__doc__ == doc

    wr.__doc__ = "changed"
    assert wr.__doc__ == "changed"

    # test reference
    assert wr.__reference__ == ref

    with pytest.raises(AttributeError):
        wr.__reference__ = "changed"

    # test meta
    assert isinstance(wr.meta, OrderedDict)
    assert wr.meta["a"] == "A"
    assert wr.meta["b"] == "B"

    assert wr.meta["reference"] == ref

    wr.meta["reference"] = "changed"
    assert wr.meta["reference"] == "changed"

    with pytest.raises(KeyError):  # it shouldn't have this
        assert wr.meta["c"]

    # --------------------------------------------------
    # TODO test something with existing meta

    return


# /def

###############################################################################
# MetaDataBase


def test_MetaDataBase_raises():
    """Test :class:`~utilipy.utils.collections.MetaDataBase` Exceptions."""
    with pytest.raises(TypeError):
        collections.MetaDataBase()  # MetaDataBase must be subclassed


# /def


def test_MetaDataBase():
    """Test :class:`~utilipy.utils.collections.MetaDataBase`."""
    # arguments into reference class
    meta = OrderedDict([("m1", "meta 1"), ("m2", "meta 2")])

    class ChildClass(collections.MetaDataBase):
        pass

    mdb = ChildClass(**meta)  # has reference in meta

    assert mdb._meta == meta
    assert mdb.meta == meta
    assert mdb.meta["m1"] == meta["m1"]


# /def


###############################################################################
# ReferenceBase


def test_ReferenceBase_raises():
    """Test :class:`~utilipy.utils.collections.ReferenceBase` Exceptions."""
    with pytest.raises(TypeError):
        collections.ReferenceBase()  # ReferenceBase must be subclassed


# /def


def test_ReferenceBase():
    """Test :class:`~utilipy.utils.collections.ReferenceBase`."""
    # arguments into reference class
    meta = OrderedDict([("m1", "meta 1"), ("m2", "meta 2")])
    reference = "the reference"

    class ChildClass(collections.ReferenceBase):
        pass

    mdb = ChildClass(reference=reference, **meta)  # has reference in meta

    # add reference into meta for comparisons
    meta["reference"] = reference

    assert mdb._meta == meta
    assert mdb.meta == meta
    assert mdb.meta["reference"] == meta["reference"] == reference
    assert mdb.meta["m1"] == meta["m1"]

    # test method
    assert mdb.__reference__ == reference


# /def


###############################################################################
# ObjDict


def test_ObjDict_creation():
    """Test creation of :class:`~utilipy.utils.collections.ObjDict`."""
    od = ObjDict(name="name", a=1, b=None, c=ObjDict("inner"))

    assert od.name == "name"
    assert od.a == 1
    assert od.b is None
    assert isinstance(od.c, ObjDict) & (od.c.name == "inner")

    return


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_getitem():
    """Test :func:`~utilipy.utils.collections.ObjDict.__getitem__`.

    - `name` is special and can only be accessed by .name
    - single item access
    - multiple item access
    - as_generator
    - getitem method

    """
    od = ObjDict(name="name", a=1, b=None, c=ObjDict("inner"))

    # name
    assert od.name == "name"

    # ------------------------------------------------------------------------
    # [] syntax

    # singular
    assert od["a"] == 1
    assert od["b"] is None
    assert isinstance(od["c"], ObjDict) & (od["c"].name == "inner")

    # plural
    assert od["a", "b"] == [1, None]
    assert od["c",] == [  # noqa
        ObjDict("inner"),
    ]

    # ------------------------------------------------------------------------
    # __getitem__

    assert od.__getitem__("a") == 1
    assert od.__getitem__("b") is None
    assert od.__getitem__("c").name == "inner"

    # plural
    assert od.__getitem__(["a", "b"]) == [1, None]
    assert od.__getitem__(["c",]) == [  # noqa
        ObjDict("inner"),
    ]

    # singular as_generator, should just be the object
    assert od.__getitem__("a", as_generator=True) == 1
    assert od.__getitem__("b", as_generator=True) is None
    assert od.__getitem__("c", as_generator=True).name == "inner"

    # plural as_generator
    iterable = od.__getitem__(("a", "b", "c"), as_generator=True)
    assert iterable.__next__() == 1
    assert iterable.__next__() is None
    assert iterable.__next__().name == "inner"

    # ------------------------------------------------------------------------
    # getitem

    assert od.getitem("a") == 1
    assert od.getitem("b") is None
    assert od.getitem("c").name == "inner"

    # plural
    assert od.getitem(["a", "b"]) == [1, None]
    assert od.getitem(["c",]) == [  # noqa
        ObjDict("inner"),
    ]

    # singular as_generator, should just be the object
    assert od.getitem("a", as_generator=True) == 1
    assert od.getitem("b", as_generator=True) is None
    assert od.getitem("c", as_generator=True).name == "inner"

    # plural as_generator
    iterable = od.getitem(("a", "b", "c"), as_generator=True)
    assert iterable.__next__() == 1
    assert iterable.__next__() is None
    assert iterable.__next__().name == "inner"

    return


# /def


# -----------------------------------------------------------------------------


def test_ObjDict__setitem__():
    """Test :func:`~utilipy.utils.collections.ObjDict.__setitem__`."""
    od = ObjDict(name="name", a=1, b=None, c=ObjDict("inner"))

    # setitem
    od["added"] = "added"
    assert od["added"] == "added"

    # __setitem__
    od.__setitem__("added2", "added2")
    assert od["added2"] == "added2"

    return


# /def


# -----------------------------------------------------------------------------


def test_ObjDict__getattr__():
    """Test :func:`~utilipy.utils.collections.ObjDict.__getattr__`."""
    _c = ObjDict("inner", a="A")
    od = ObjDict(name="name", a=1, b=None, c=_c)

    assert od.__getattr__("a") == 1
    assert od.__getattr__("b") is None
    assert od.__getattr__("c").__getattr__("a") == "A"

    with pytest.raises(KeyError):
        assert od.__getattr__("name")  # only by __getattribute__

    return


# /def


# -----------------------------------------------------------------------------


def test_ObjDict__setattr__():
    """Test :func:`~utilipy.utils.collections.ObjDict.__setattr__`."""
    # make ObjDict
    od = ObjDict()
    od.__setattr__("a", 1)

    # test
    assert od.__getattr__("a") == 1

    return


# /def

# -----------------------------------------------------------------------------


def test_ObjDict__repr__():
    """Test :func:`~utilipy.utils.collections.ObjDict.__repr__`."""
    # --------------------------
    # without a name
    od = ObjDict(a=2)
    assert od.__repr__() == OrderedDict({"a": 2}).__repr__().replace(
        "OrderedDict", "ObjDict"
    )

    # --------------------------
    # with a name
    od = ObjDict("name", a=2)
    assert od.__repr__() == od.name + OrderedDict({"a": 2}).__repr__().replace(
        "OrderedDict", ""
    )
    return


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_values():
    """Test :func:`~utilipy.utils.collections.ObjDict.values`."""
    _c = ObjDict("inner", a="A")
    od = ObjDict(name="name", a=1, b=None, c=_c)

    # test standard (without keys)
    assert list(od.values()) == [1, None, _c]

    # test with keys
    assert list(od.values("a", "b")) == [1, None]

    return


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_items():
    """Test :func:`~utilipy.utils.collections.ObjDict.items`."""
    _c = ObjDict("inner", a="A")
    od = ObjDict(name="name", a=1, b=None, c=_c)

    # test standard (without keys)
    assert list(od.items()) == [
        ("a", 1),
        ("b", None),
        ("c", _c),
    ]

    # test when provide keys
    assert list(od.items("a", "b")) == [
        ("a", 1),
        ("b", None),
    ]

    return


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_fromkeys():
    """Test :func:`~utilipy.utils.collections.ObjDict.fromkeys`."""
    od = ObjDict(name="name", a=1, b=None, c="c")

    # test standard (without keys)
    assert od.fromkeys() == od

    # test standard, without keys, but name
    # shouldn't do anything
    assert od.fromkeys(name="new") == od

    # test with keys provided
    assert od.fromkeys(["a", "b"]) == ObjDict(name="name subset", a=1, b=None)
    assert od.fromkeys(["a", "b"], "Name") == ObjDict(name="Name", a=1, b=None)

    return


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_keyslist():
    """Test :func:`~utilipy.utils.collections.ObjDict.keyslist`."""
    od = ObjDict(name="name", a=1, b=None, c="c")

    assert set(od.keyslist()) == {"a", "b", "c"}


# /def


# -----------------------------------------------------------------------------


def test_ObjDict__reduce__():
    """Test :func:`~utilipy.utils.collections.ObjDict.__reduce__`."""
    od = ObjDict(name="name", a=1, b=None, c="c")

    reduced = od.__reduce__()

    assert reduced[0] == od.__class__
    assert reduced[1] == (od.name,)
    assert reduced[2] == OrderedDict(od.items())


# /def


# -----------------------------------------------------------------------------


def test_ObjDict__setstate__():
    """Test :func:`~utilipy.utils.collections.ObjDict.__setstate__`."""
    od = ObjDict(name="name", a=1, b=None, c="c")

    new_od = ObjDict()
    new_od.__setstate__(od)

    assert new_od == od


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_dump():
    """Test :func:`~utilipy.utils.collections.ObjDict.dump`."""
    od = ObjDict(name="name", a=1, b=None, c="c")

    with tempfile.TemporaryDirectory() as tempdir:
        tempath = os.path.join(tempdir, "temp.pkl")
        od.dump(
            tempath,
            protocol=None,
            fopt="b",
            fix_imports=True,
        )

    # /with


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_save():
    """Test :func:`~utilipy.utils.collections.ObjDict.save`."""
    od = ObjDict(name="name", a=1, b=None, c="c")

    with tempfile.TemporaryDirectory() as tempdir:
        tempath = os.path.join(tempdir, "temp.pkl")
        od.save(
            tempath,
            protocol=None,
            fopt="b",
            fix_imports=True,
        )

    # /with


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_load():
    """Test :func:`~utilipy.utils.collections.ObjDict.load`."""
    od = ObjDict(name="name", a=1, b=None, c="c")

    with tempfile.TemporaryDirectory() as tempdir:
        tempath = os.path.join(tempdir, "temp.pkl")
        od.dump(
            tempath,
            protocol=None,
            fopt="b",
            fix_imports=True,
        )

        new_od = ObjDict.load(
            tempath,
            fopt="b",
            fix_imports=True,
            encoding="ASCII",
            errors="strict",
        )

    # /with

    # test equality between new and old
    assert new_od.name == od.name
    assert new_od.items() == od.items()


# /def


# -----------------------------------------------------------------------------


def test_ObjDict_print():  # TODO better test
    """Test :func:`~utilipy.utils.collections.ObjDict.print`.

    Right now, only testing that it doesn't error.

    """
    od = ObjDict(name="name", a=1, b=None, c="c")

    od.print()


# /def


##########################################################################


###############################################################################
# END
